#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This module is deprecated. It will be removed in one of the next versions of smartg.
# All the classes/functions are now available in the geoclide package.

from smartg.geometry import Vector, Point, Normal, Ray, BBox
from smartg.geometry import Normalize
import numpy as np
from numpy.linalg import inv
from warnings import warn

#####################################################################################
class Transform(object):
    '''
    Creation of the class Transform
    '''

    def __init__(self, m = None, mInv = None):
        warn_message = 'Using Transform from smartg.transform is depracated as of smartg v1.1.0. ' + \
                       'Use geoclide package instead.\n For example: import geoclide as gc; gc.Transform().'
        warn(warn_message, DeprecationWarning)
        if m is None : m = np.identity(4)
        if (isinstance(m, Transform)):
            self.m = m.m
            self.mInv = m.mInv
        else:
            self.m = m
            if (mInv is None):
                self.mInv = inv(self.m)
            else:
                self.mInv = mInv

    def __eq__(self, t):
        if isinstance(t, Transform):
            self.m = t.m
            self.mInv = t.mInv

    def inverse(self, t):
        return Transform(t.mInv, t.m)

    def isIdentity(self):
        return (self.m[0,0] == 1) and (self.m[0,1] == 0) and (self.m[0,2] == 0) and \
            (self.m[0,3] == 0) and (self.m[1,0] == 0) and (self.m[1,1] == 1) and \
            (self.m[1,2] == 0) and (self.m[1,3] == 0) and (self.m[2,0] == 0) and \
            (self.m[2,1] == 0) and (self.m[2,2] == 1) and (self.m[2,3] == 0) and \
            (self.m[3,0] == 0) and (self.m[3,1] == 0) and (self.m[3,2] == 0) and \
            (self.m[3,3] == 1)

    def translate(self, v):
        myM = np.identity(4)
        myM[0,3] = v.x; myM[1,3] = v.y; myM[2,3] = v.z
        myMInv = np.identity(4)
        myMInv[0,3] = (v.x)*-1; myMInv[1,3] = (v.y)*-1; myMInv[2,3] = (v.z)*-1
        return Transform(myM, myMInv)

    def scale(self, x, y, z):
        myM = np.identity(4)
        myM[0,0] = x; myM[1,1] = y; myM[2,2] = z
        myMInv = np.identity(4)
        myMInv[0,0] = 1./x; myMInv[1,1] = 1./y; myMInv[2,2] = 1./z
        return Transform(myM, myMInv)

    def rotateX(self, angle):
        sin_t = np.sin(angle*(np.pi / 180.))
        cos_t = np.cos(angle*(np.pi / 180.))
        myM = np.identity(4)
        myM[1,1] = cos_t; myM[1,2] = -1.*sin_t
        myM[2,1] = sin_t; myM[2,2] = cos_t
        return Transform(myM, np.transpose(myM))

    def rotateY(self, angle):
        sin_t = np.sin(angle*(np.pi / 180.))
        cos_t = np.cos(angle*(np.pi / 180.))
        myM = np.identity(4)
        myM[0,0] = cos_t; myM[2,0] = -1.*sin_t
        myM[0,2] = sin_t; myM[2,2] = cos_t
        return Transform(myM, np.transpose(myM))

    def rotateZ(self, angle):
        sin_t = np.sin(angle*(np.pi / 180.))
        cos_t = np.cos(angle*(np.pi / 180.))
        myM = np.identity(4)
        myM[0,0] = cos_t; myM[0,1] = -1.*sin_t
        myM[1,0] = sin_t; myM[1,1] = cos_t
        return Transform(myM, np.transpose(myM))

    def rotate(self, angle, axis):
        a = Vector(Normalize(axis))
        s = np.sin(angle*(np.pi / 180.))
        c = np.cos(angle*(np.pi / 180.))
        myM = np.identity(4)

        myM[0,0] = a.x*a.x+(1-a.x*a.x)*c
        myM[0,1] = a.x*a.y*(1-c)-a.z*s
        myM[0,2] = a.x*a.z*(1-c)+a.y*s

        myM[1,0] = a.x*a.y*(1-c)+a.z*s
        myM[1,1] = a.y*a.y+(1-a.y*a.y)*c
        myM[1,2] = a.y*a.z*(1-c)-a.x*s

        myM[2,0] = a.x*a.z*(1-c)-a.y*s
        myM[2,1] = a.y*a.z*(1-c)+a.x*s
        myM[2,2] = a.z*a.z+(1-a.z*a.z)*c
        
        return Transform(myM, np.transpose(myM))

    def __mul__(self, T): 
        if (isinstance(T, Transform)):
            return Transform(np.dot(self.m, T.m), np.dot(T.mInv, self.mInv))
        else:
            raise NameError('mul accepted only with Transform')

    def __getitem__(self, c):
        if isinstance(c, Point):
            xp = self.m[0,0]*c.x + self.m[0,1]*c.y + self.m[0,2]*c.z + self.m[0,3]
            yp = self.m[1,0]*c.x + self.m[1,1]*c.y + self.m[1,2]*c.z + self.m[1,3]
            zp = self.m[2,0]*c.x + self.m[2,1]*c.y + self.m[2,2]*c.z + self.m[2,3]
            wp = self.m[3,0]*c.x + self.m[3,1]*c.y + self.m[3,2]*c.z + self.m[3,3]
            if (wp == 1):
                return Point(xp, yp, zp)
            else: 
                return Point(xp, yp, zp)/wp
        elif isinstance(c, Vector):
            xv = self.m[0,0]*c.x + self.m[0,1]*c.y + self.m[0,2]*c.z
            yv = self.m[1,0]*c.x + self.m[1,1]*c.y + self.m[1,2]*c.z
            zv = self.m[2,0]*c.x + self.m[2,1]*c.y + self.m[2,2]*c.z
            return Vector(xv, yv, zv)
        elif isinstance(c, Normal):
            xn = self.mInv[0,0]*c.x + self.mInv[1,0]*c.y + self.mInv[2,0]*c.z
            yn = self.mInv[0,1]*c.x + self.mInv[1,1]*c.y + self.mInv[2,1]*c.z
            zn = self.mInv[0,2]*c.x + self.mInv[1,2]*c.y + self.mInv[2,2]*c.z
            return Normal(xn, yn, zn)
        elif isinstance(c, Ray):
            R = Ray(c.o, c.d)
            R.o = self[R.o]
            R.d = self[R.d]
            return R
        elif isinstance(c, BBox):
            P1 = self[Point(c.pMin.x, c.pMin.y, c.pMin.z)]
            V = self[Vector(c.pMax.x-c.pMin.x, c.pMax.y-c.pMin.y, c.pMax.z-c.pMin.z)]
            P2 = P1 + V
            # print "V =", V, "P1=", P1, "P2=", P2
            ret = BBox(P1, P2)
            # print "minB =", ret.pMin, "maxB =", ret.pMax
            # ret = BBox(self[Point(c.pMin.x, c.pMin.y, c.pMin.z)]) 
            # ret = ret.union(ret, self[Point(c.pMax.x, c.pMin.y, c.pMin.z)])
            # ret = ret.union(ret, self[Point(c.pMin.x, c.pMax.y, c.pMin.z)])
            # ret = ret.union(ret, self[Point(c.pMin.x, c.pMin.y, c.pMax.z)])
            # ret = ret.union(ret, self[Point(c.pMin.x, c.pMax.y, c.pMax.z)])
            # ret = ret.union(ret, self[Point(c.pMax.x, c.pMax.y, c.pMin.z)])
            # ret = ret.union(ret, self[Point(c.pMax.x, c.pMin.y, c.pMax.z)])
            # ret = ret.union(ret, self[Point(c.pMax.x, c.pMax.y, c.pMax.z)])
            return ret
        else:
            raise NameError('Unkonwn type for transformations')
#####################################################################################

#####################################################################################
def Aff(Mat, myString = None):
    if (myString is None):
        myString = 'My Matrix'
    myString2 = ' '
    for i in range(0, len(myString)-1):
        myString2 = myString2 + ' '

    print(myString + ' = (' + str(Mat[0,0]) + ', ' + str(Mat[0,1]) + ', ' \
        + str(Mat[0,2]) + ', ' + str(Mat[0,3])+ ')')
    print(myString2 + '   (' + str(Mat[1,0]) + ', ' + str(Mat[1,1]) + ', ' \
        + str(Mat[1,2]) + ', ' + str(Mat[1,3])+ ')')
    print(myString2 + '   (' + str(Mat[2,0]) + ', ' + str(Mat[2,1]) + ', ' \
        + str(Mat[2,2]) + ', ' + str(Mat[2,3])+ ')')
    print(myString2 + '   (' + str(Mat[3,0]) + ', ' + str(Mat[3,1]) + ', ' \
        + str(Mat[3,2]) + ', ' + str(Mat[3,3])+ ')')
#####################################################################################

##################################################################################### 
if __name__ == '__main__':

    A = np.array([[1, 0, 0, 2], [0, 1, 0, 2], [0, 0, 1, 2], [0, 0, 0, 1]])
    P1 = Point(10, 0, 2)
    V1 = Vector(1, 2, 3)
    R1 = Ray(P1, V1)

    print("P1 =", P1, "V1 =", V1, "R1[1] =", R1[1])

    MyT=Transform()
    Aff(MyT.m, 'Mat')
    Aff(MyT.mInv, 'InvMat')
    Trans = MyT.translate(Vector(10, 0, 0))
    Tinv = Trans.inverse(Trans)
    P2 = Tinv[P1]
    V2 = Tinv[V1]
    R2 = Tinv[R1]
    
    print("translated point =", P2, ", vector =", V2, ", ray =", R2[1])

    muSca = MyT.scale(2,2,2)
    muScaInv = muSca.inverse(muSca)
    TS = Trans*muSca #Transform(np.dot(Trans.m, muSca.m), np.dot(muSca.mInv, Trans.mInv))
    TSInv = TS.inverse(TS)

    P3 = muScaInv[P2]
    V3 = muScaInv[V2]
    R3 = muScaInv[R2]

    print("trans + scale : point =", P3, ", vector =", V3, ", ray =", R3[1])

    P4 = TSInv[P1]
    V4 = TSInv[V1]
    R4 = TSInv[R1]

    print("trans + scale 2: point =", P4, ", vector =", V4, ", ray =", R4[1])
    # Aff(MyT.mInv, 'MatInv')
    # MyT = MyT.Inverse(MyT)
    # Aff(MyT.m, 'Mat')
    # Aff(MyT.mInv, 'MatInv')



#!/usr/bin/env python
# encoding: utf-8


from __future__ import print_function, division, absolute_import

import warnings
warnings.simplefilter("ignore",DeprecationWarning)
from pylab import figure, subplot2grid, tight_layout, setp, subplots, xlabel, ylabel, FormatStrFormatter
import numpy as np
np.seterr(invalid='ignore', divide='ignore') # ignore division by zero errors
from luts.luts import plot_polar, transect2D, Idx
from smartg.atmosphere import diff1
from smartg.water import diff2


def mdesc(desc, logI=False):
    sep1=desc.find('_')
    sep2=desc.find('(')
    sep3=desc.find(')')
    if sep1 == 1 : stokes=desc[0:1]
    elif sep1 == 2 : stokes=desc[sep1-2:sep1]
    elif sep1 == 4 : stokes=desc[sep1-4:sep1]
    else : stokes=desc[0:sep1]
    dir=desc[sep1+1:sep2-1]

    if logI and stokes=='I':
        pref=r'$log_{10} '
    else:
        pref=r'$'
        
    if dir == 'up':
        return pref + stokes + r'^{\uparrow}' + '_{'+desc[sep2+1:sep3]+'}' + desc[sep3+1:] +'$'
    else:
        return pref + stokes + r'^{\downarrow}' + '_{'+desc[sep2+1:sep3]+'}' + desc[sep3+1:] +'$'
    

def smartg_view(mlut, logI=False, QU=False, Circ=False, full=False, field='up (TOA)', prefix='', ind=[0], cmap=None, fig=None, subdict=None,
        Imin=None, Imax=None, Pmin=0, Pmax=100):
    '''
    visualization of a smartg MLUT

    Options:
        logI: shows log10 of I
        Circ: shows Circular polarization 
        QU:  shows Q U and DoP
        field: level of output
        prefix: eventually a prefix for field
        ind: list of indices of azimutal planes
        full: shows all
        cmap: color map
        fig: already existing figure
        subdict: dictionnary of LUT subsetter (see LUT class , sub() method)
        Imin: min value of I
        Imax: max value of I
        Pmin: min value of DOP
        Pmax: max value of DOP

    Outputs:
    if full is False, it returns 1 figure
    if full is True,  it returns 2 figures
    '''

    I = mlut[prefix+'I_' + field]
    Q = mlut[prefix+'Q_' + field]
    U = mlut[prefix+'U_' + field]
    V = mlut[prefix+'V_' + field]

    if subdict is not None :
        I = I.sub(d=subdict)
        Q = Q.sub(d=subdict)
        U = U.sub(d=subdict)
        V = V.sub(d=subdict)

    # Linearly polarized reflectance
    IPL = (Q*Q + U*U).apply(np.sqrt, 'Lin. Pol. ref.')
    
    # Polarized reflectance
    IP = (Q*Q + U*U + V*V).apply(np.sqrt, 'Pol. ref.')

    # Degree of Linear Polarization (%)
    DoLP = 100*IPL/I
    DoLP.desc = prefix+r'$DoLP$'
    
    # Angle of Linear Polarization (deg)
    AoLP = (U/Q)
    AoLP.apply(np.arctan)*90/np.pi
    AoLP.desc = prefix+r'$AoLP$'
    
    # Degree of Circular Polarization (%)
    DoCP = 100*V.apply(abs)/I
    DoCP.desc = prefix+r'$DoCP$'

    # Degree of Polarization (%)
    DoP = 100*IP/I
    DoP.desc = prefix+r'$DoP\,(\%)$'

    if not full:
        if QU:
            if fig is None: fig = figure(figsize=(9, 9))
            if logI:
                lI=I.apply(np.log10)
                lI.desc = mdesc(I.desc, logI=logI)
                plot_polar(lI,  index=ind, rect=421, sub=423, fig=fig, cmap=cmap, vmin=Imin, vmax=Imax)
            else:
                I.desc = mdesc(I.desc)
                plot_polar(I,  index=ind, rect=421, sub=423, fig=fig, cmap=cmap, vmin=Imin, vmax=Imax)
            Q.desc = mdesc(Q.desc)
            U.desc = mdesc(U.desc)
            plot_polar(Q,  index=ind, rect=422, sub=424, fig=fig, cmap=cmap)
            plot_polar(U,  index=ind, rect=425, sub=427, fig=fig, cmap=cmap)
            if Circ:
                V.desc = mdesc(V.desc)
                plot_polar(V, index=ind, rect=426, sub=428, fig=fig, cmap=cmap)
            else:
                plot_polar(DoP, index=ind, rect=426, sub=428, fig=fig, vmin=Pmin, vmax=Pmax, cmap=cmap)
        else:
            # show only I and PR
            if fig is None: fig = figure(figsize=(9, 4.5))
            if logI:
                lI=I.apply(np.log10)
                lI.desc = mdesc(I.desc, logI=logI)
                plot_polar(lI,  index=ind, rect=221, sub=223, fig=fig, cmap=cmap, vmin=Imin, vmax=Imax)
            else:
                I.desc = mdesc(I.desc)
                plot_polar(I,  index=ind, rect=221, sub=223, fig=fig, cmap=cmap, vmin=Imin, vmax=Imax)

            if Circ:
                plot_polar(DoCP, index=ind, rect=222, sub=224, fig=fig, vmin=0, vmax=Pmax, cmap=cmap)
            else:
                plot_polar(DoP, index=ind, rect=222, sub=224, fig=fig, vmin=Pmin, vmax=Pmax, cmap=cmap)

        return fig


    else:
        # full plots
        fig1 = figure(figsize=(16, 4))
        lI=I.apply(np.log10)
        lI.desc = mdesc(I.desc,logI=True)
        I.desc = mdesc(I.desc)
        Q.desc = mdesc(Q.desc)
        U.desc = mdesc(U.desc)
        V.desc = mdesc(V.desc)
        plot_polar(I,  index=ind, rect=241, sub=245, fig=fig1, cmap=cmap, vmin=Imin, vmax=Imax)
        plot_polar(Q,  index=ind, rect=242, sub=246, fig=fig1, cmap=cmap)
        plot_polar(U,  index=ind, rect=243, sub=247, fig=fig1, cmap=cmap)
        plot_polar(V,  index=ind, rect=244, sub=248, fig=fig1, cmap=cmap)
        
        fig2 = figure(figsize=(16, 4))
        Q.desc = mdesc(Q.desc)
        U.desc = mdesc(U.desc)
        V.desc = mdesc(V.desc)
        plot_polar(lI,  index=ind, rect=241, sub=245, fig=fig2, cmap=cmap)
        plot_polar(DoLP,  index=ind, rect=242, sub=246, fig=fig2, vmin=Pmin, vmax=Pmax, cmap=cmap)
        plot_polar(DoCP,  index=ind, rect=243, sub=247, fig=fig2, vmin=Pmin, vmax=Pmax, cmap=cmap)
        plot_polar(DoP,  index=ind, rect=244, sub=248, fig=fig2, vmin=Pmin, vmax=Pmax, cmap=cmap)
        #plot_polar(AoLP,  index=ind, rect='244', sub='248', fig=fig2, vmin=-180, vmax=180, cmap=cmap)

        return fig1, fig2

def transect_view(mlut, logI=False, QU=False, Circ=False, full=False, field='up (TOA)', prefix='', ind=[0], fig=None, color='k', subdict=None, 
         **kwargs):
    '''
    visualization of a smartg MLUT

    Options:
        logI: shows log10 of I
        Circ: shows Circular polarization 
        QU:  shows Q U and DoP
        field: level of output
        ind: list of indices of azimutal planes
        full: shows all
        color: color of the transect
        subdict: dictionnary of LUT subsetter (see LUT class , sub() method)

    Outputs:
    if full is False, it returns 1 figure
    if full is True,  it returns 2 figures
    '''

    I = mlut[prefix+'I_' + field]
    Q = mlut[prefix+'Q_' + field]
    U = mlut[prefix+'U_' + field]
    V = mlut[prefix+'V_' + field]

    if subdict is not None :
        I = I.sub(d=subdict)
        Q = Q.sub(d=subdict)
        U = U.sub(d=subdict)
        V = V.sub(d=subdict)

    # Linearly polarized reflectance
    IPL = (Q*Q + U*U).apply(np.sqrt, 'Lin. Pol. ref.')
    
    # Polarized reflectance
    IP = (Q*Q + U*U +V*V).apply(np.sqrt, 'Pol. ref.')

    # Degree of Linear Polarization (%)
    DoLP = 100*IPL/I
    DoLP.desc = prefix+r'$DoLP$'
    
    # Angle of Linear Polarization (deg)
    AoLP = (U/Q)
    AoLP.apply(np.arctan)*90/np.pi
    AoLP.desc = prefix+r'$AoLP$'
    
    # Degree of Circular Polarization (%)
    DoCP = 100*V.apply(abs)/I
    DoCP.desc = prefix+r'$DoCP$'

    # Degree of Polarization (%)
    DoP = 100*IP/I
    DoP.desc = prefix+r'$DoP$'

    if not full:
        if QU:
            if fig is None: fig = figure(figsize=(8, 8))
            if logI:
                lI=I.apply(np.log10)
                lI.desc = mdesc(I.desc, logI=logI)
                transect2D(lI,  index=ind, sub=221, fig=fig, color=color,  **kwargs)
            else:
                I.desc = mdesc(I.desc)
                transect2D(I,  index=ind, sub=221, fig=fig, color=color,   **kwargs)
            Q.desc = mdesc(Q.desc)
            U.desc = mdesc(U.desc)
            transect2D(Q,  index=ind, sub=222, fig=fig, color=color, **kwargs)
            transect2D(U,  index=ind, sub=223, fig=fig, color=color, **kwargs)
            if Circ:
                V.desc = mdesc(V.desc)
                transect2D(V, index=ind, sub=224, fig=fig, color=color, **kwargs)
            else:
                transect2D(DoP, index=ind, sub=224, fig=fig,  color=color, percent=True, **kwargs)
        else:
            # show only I and PR
            if fig is None: fig = figure(figsize=(8, 4))
            if logI:
                lI=I.apply(np.log10)
                lI.desc = mdesc(I.desc, logI=logI)
                transect2D(lI,  index=ind, sub=121, fig=fig, color=color,   **kwargs)
            else:
                I.desc = mdesc(I.desc)
                transect2D(I,  index=ind, sub=121, fig=fig, color=color,  **kwargs)

            if Circ:
                transect2D(DoCP, index=ind, sub=122, fig=fig,  color=color, percent=True, **kwargs)
            else:
                transect2D(DoP, index=ind, sub=122, fig=fig, color=color, percent=True, **kwargs)

        return fig


    else:
        # full plots
        if fig is None: 
            fig1 = figure(figsize=(16, 4))
            fig2 = figure(figsize=(16, 4))
        else : fig1,fig2 = fig
        lI=I.apply(np.log10)
        lI.desc = mdesc(I.desc,logI=True)
        I.desc = mdesc(I.desc)
        Q.desc = mdesc(Q.desc)
        U.desc = mdesc(U.desc)
        V.desc = mdesc(V.desc)
        transect2D(I,  index=ind,  sub=141, fig=fig1, color=color,  **kwargs)
        transect2D(Q,  index=ind,  sub=142, fig=fig1, color=color, **kwargs)
        transect2D(U,  index=ind, sub=143, fig=fig1, color=color, **kwargs)
        transect2D(V,  index=ind, sub=144, fig=fig1, color=color, **kwargs)
        
        Q.desc = mdesc(Q.desc)
        U.desc = mdesc(U.desc)
        V.desc = mdesc(V.desc)
        transect2D(lI,  index=ind, sub=141,fig=fig2, color=color, **kwargs)
        transect2D(DoLP,  index=ind, sub=142, fig=fig2, color=color, percent=True, **kwargs)
        transect2D(DoCP,  index=ind, sub=143, fig=fig2, color=color, percent=True, **kwargs)
        transect2D(DoP,  index=ind,  sub=144, fig=fig2, color=color, percent=True, **kwargs)
        #transect2D(AoLP, index=ind,  sub=144, fig=fig2, color=color, **kwargs)

        return fig1, fig2

def spectrum(lut, vmin=None, vmax=None, sub='111', fig=None, color='k', percent=False, fmt='-'):
    '''
    spectrum of a 1D LUT

    lut:  1D look-up table to display
            with axis 'wavelength'
    vmin, vmax: range of values
                default None: determine min/max from values
    fig : destination figure. If None (default), create a new figure.
    color : color of the transect
    percent: if True set scale to 0 to 100%
    '''
    from pylab import figure

    assert 'wavelength' in lut.names

    if fig is None:
        fig = figure(figsize=(4.5, 2.5))

    ax1   = lut.axes[0]
    data  = lut.data

    if vmin is None:
        vmin = np.amin(lut.data[~np.isnan(lut.data)])
    if vmax is None:
        vmax = np.amax(lut.data[~np.isnan(lut.data)])
    if vmin == vmax:
        vmin -= 0.001
        vmax += 0.001
    if vmin > vmax: vmin, vmax = vmax, vmin
    if percent:
        vmin=0.
        vmax=100.

    ax1_min = np.amin(ax1)
    ax1_max = np.amax(ax1)
    #

    ax_cart = fig.add_subplot(sub)
    ax_cart.grid(True)

    ax_cart.set_xlim(ax1_min, ax1_max)
    ax_cart.set_ylim(vmin, vmax)
    ax_cart.ticklabel_format(axis='y', style='sci', scilimits=(-2,2))
    #ax_cart.ticklabel_format(axis='x', style='sci', scilimits=(-2,2))
    ax_cart.grid(True)
    ax_cart.set_xlabel(r'$\lambda$ (nm)')

    #
    # plot
    #
    ax_cart.plot(ax1 , data[:], fmt, color=color)

    if lut.desc is not None:
        ax_cart.set_title(lut.desc)


def spectrum_view(mlut, logI=False, QU=False, Circ=False, full=False, field='up (TOA)', prefix='', fig=None, color='k', subdict=None, 
         **kwargs):
    '''
    visualization of a smartg MLUT

    Options:
        logI: shows log10 of I
        Circ: shows Circular polarization 
        QU:  shows Q U and DoP
        field: level of output
        full: shows all
        color: color of the transect
        subdict: dictionnary of LUT subsetter (see LUT class , sub() method)

    Outputs:
    if full is False, it returns 1 figure
    if full is True,  it returns 2 figures
    '''

    I = mlut[prefix+'I_' + field]
    Q = mlut[prefix+'Q_' + field]
    U = mlut[prefix+'U_' + field]
    V = mlut[prefix+'V_' + field]

    if subdict is not None :
        I = I.sub(d=subdict)
        Q = Q.sub(d=subdict)
        U = U.sub(d=subdict)
        V = V.sub(d=subdict)

    # Linearly polarized reflectance
    IPL = (Q*Q + U*U).apply(np.sqrt, 'Lin. Pol. ref.')
    
    # Polarized reflectance
    IP = (Q*Q + U*U +V*V).apply(np.sqrt, 'Pol. ref.')

    # Degree of Linear Polarization (%)
    DoLP = 100*IPL/I
    DoLP.desc = prefix+r'$DoLP$'
    
    # Angle of Linear Polarization (deg)
    AoLP = (U/Q)
    AoLP.apply(np.arctan)*90/np.pi
    AoLP.desc = prefix+r'$AoLP$'
    
    # Degree of Circular Polarization (%)
    DoCP = 100*V.apply(abs)/I
    DoCP.desc = prefix+r'$DoCP$'

    # Degree of Polarization (%)
    DoP = 100*IP/I
    DoP.desc = prefix+r'$DoP$'

    if not full:
        if QU:
            if fig is None: fig = figure(figsize=(8, 8))
            if logI:
                lI=I.apply(np.log10)
                lI.desc = mdesc(I.desc, logI=logI)
                spectrum(lI, sub=221, fig=fig, color=color,  **kwargs)
            else:
                I.desc = mdesc(I.desc)
                spectrum(I,  sub=221, fig=fig, color=color,   **kwargs)
            Q.desc = mdesc(Q.desc)
            U.desc = mdesc(U.desc)
            spectrum(Q, sub=222, fig=fig, color=color, **kwargs)
            spectrum(U, sub=223, fig=fig, color=color, **kwargs)
            if Circ:
                V.desc = mdesc(V.desc)
                spectrum(V, sub=224, fig=fig, color=color, **kwargs)
            else:
                spectrum(DoP, sub=224, fig=fig,  color=color, percent=True, **kwargs)
        else:
            # show only I and PR
            if fig is None: fig = figure(figsize=(8, 4))
            if logI:
                lI=I.apply(np.log10)
                lI.desc = mdesc(I.desc, logI=logI)
                spectrum(lI, sub=121, fig=fig, color=color,   **kwargs)
            else:
                I.desc = mdesc(I.desc)
                spectrum(I, sub=121, fig=fig, color=color,  **kwargs)

            if Circ:
                spectrum(DoCP, sub=122, fig=fig,  color=color, percent=True, **kwargs)
            else:
                spectrum(DoP, sub=122, fig=fig, color=color, percent=True, **kwargs)

        return fig

    else:
        # full plots
        if fig is None: 
            fig1 = figure(figsize=(16, 4))
            fig2 = figure(figsize=(16, 4))
        else : fig1,fig2 = fig
        lI=I.apply(np.log10)
        lI.desc = mdesc(I.desc,logI=True)
        I.desc = mdesc(I.desc)
        Q.desc = mdesc(Q.desc)
        U.desc = mdesc(U.desc)
        V.desc = mdesc(V.desc)
        spectrum(I, sub=141, fig=fig1, color=color,  **kwargs)
        spectrum(Q, sub=142, fig=fig1, color=color, **kwargs)
        spectrum(U, sub=143, fig=fig1, color=color, **kwargs)
        spectrum(V, sub=144, fig=fig1, color=color, **kwargs)
        
        spectrum(lI, sub=141, fig=fig2, color=color, **kwargs)
        spectrum(DoLP, sub=142, fig=fig2, color=color, percent=True, **kwargs)
        spectrum(DoCP, sub=143, fig=fig2, color=color, percent=True, **kwargs)
        spectrum(DoP, sub=144, fig=fig2, color=color, percent=True, **kwargs)
        #spectrum(AoLP, index=ind,  sub=144, fig=fig2, color=color, **kwargs)

        return fig1, fig2
        
def phase_view(mlut, ipha=None, fig=None, axarr=None, iw=0, kind='atm'):
    '''
    visualization of a smartg MLUT phase function from output

    Options:
        ipha: absolute index of the phase function coming from Profile
        fig : fig object to be created or included in
        axarr : system of axes (2,2) to be created or used
        iw : in case of multi wavelength simulation, index of wavelength to be plotted
        kind : atmopsheric 'atm' or oceanic 'oc' phase function
    '''

    nd = mlut['OD_'+kind].ndim
    if nd>1:
        wi = mlut['OD_'+kind].names.index('wavelength') # Wavelength index
        key = [slice(None)]*nd
        key[wi] = iw
        key=tuple(key)
        labw=r' at $%.1f nm$'%mlut.axes['wavelength'][iw]
    else:
        key=tuple([slice(None)])
        labw=''

    phase = mlut['phase_'+kind]
    ang = phase.axis('theta_'+kind)
    nstk = len(phase[0,:,0])
    if (axarr is None):
        if nstk == 4:
            fig, axarr = subplots(2, 2)
            fig.set_size_inches(10, 6)
        elif nstk == 6:
            fig, axarr = subplots(nrows=3, ncols=2)
            fig.set_size_inches(10, 9)
        
    if (ipha is None) : ni= np.unique(mlut['iphase_'+kind].__getitem__(key)) 
    else:ni=[ipha]
    
    for i in ni:
        if nstk == 4:
            P11 = 0.5*(phase[i,0,:]+phase[i,1,:])
            P12 = 0.5*(phase[i,0,:]-phase[i,1,:])
            P33 = phase[i,2,:]
            P43 = phase[i,3,:]
        
            if (np.max(P11[:]) > 0.) :axarr[0,0].semilogy(ang, P11,label='%3i'%i)
            axarr[0,0].set_title(r'$P_{11}$'+labw)
            axarr[0,0].grid()
            axarr[0,0].set_xlim([0,180])
            axarr[0,0].set_xticks([0,30,60,90,120,150,180])
            
            if (np.max(P11[:]) > 0.) :axarr[0,1].plot(ang, -P12/P11)
            axarr[0,1].set_title(r'-$P_{12}/P_{11}$')
            axarr[0,1].grid()
            axarr[0,1].set_xlim([0,180])
            axarr[0,1].set_xticks([0,30,60,90,120,150,180])

            
            if (np.max(P11[:]) > 0.) :axarr[1,0].plot(ang, P33/P11)
            axarr[1,0].set_title(r'$P_{33}/P_{11}$')
            axarr[1,0].grid()
            axarr[1,0].set_xlim([0,180])
            axarr[1,0].set_xlabel(r'$\theta$')
            axarr[1,0].set_xticks([0,30,60,90,120,150,180])
                    
            if (np.max(P11[:]) > 0.) :axarr[1,1].plot(ang, P43/P11)
            axarr[1,1].set_title(r'$P_{43}/P_{11}$')
            axarr[1,1].grid()
            axarr[1,1].set_xlim([0,180])
            axarr[1,1].set_xlabel(r'$\theta$')
            axarr[1,1].set_xticks([0,30,60,90,120,150,180])
        elif nstk == 6:
            F0 = phase[i,0,:] # F11
            F1 = phase[i,1,:] # F12 = F21
            F2 = phase[i,2,:] # F33
            F3 = phase[i,3,:] # F34 = -F43
            F4 = phase[i,4,:] # F22
            F5 = phase[i,5,:] # F44

            P11 = 0.5*(F0+2*F1+F4)
            P12 = 0.5*(F0-F4)
            P22 = 0.5*(F0-2*F1+F4)
            P33 = F2
            P34 = F3
            P44 = F5
        
            if (np.max(P11[:]) > 0.) :axarr[0,0].semilogy(ang, P11,label='%3i'%i)
            axarr[0,0].set_title(r'$P_{11}$'+labw)
            axarr[0,0].grid()
            axarr[0,0].set_xlim([0,180])
            axarr[0,0].set_xticks([0,30,60,90,120,150,180])
            
            if (np.max(P11[:]) > 0.) :axarr[0,1].plot(ang, -P12/P11)
            axarr[0,1].set_title(r'-$P_{12}/P_{11}$')
            axarr[0,1].grid()
            axarr[0,1].set_xlim([0,180])
            axarr[0,1].set_xticks([0,30,60,90,120,150,180])

            
            if (np.max(P11[:]) > 0.) :axarr[1,0].plot(ang, P33/P11)
            axarr[1,0].set_title(r'$P_{33}/P_{11}$')
            axarr[1,0].grid()
            axarr[1,0].set_xlim([0,180])
            axarr[1,0].set_xlabel(r'$\theta$')
            axarr[1,0].set_xticks([0,30,60,90,120,150,180])
                    
            if (np.max(P11[:]) > 0.) :axarr[1,1].plot(ang, P34/P11)
            axarr[1,1].set_title(r'$P_{34}/P_{11}$')
            axarr[1,1].grid()
            axarr[1,1].set_xlim([0,180])
            axarr[1,1].set_xlabel(r'$\theta$')
            axarr[1,1].set_xticks([0,30,60,90,120,150,180])

            if (np.max(P11[:]) > 0.) :axarr[2,0].plot(ang, P22/P11)
            axarr[2,0].set_title(r'$P_{22}/P_{11}$')
            axarr[2,0].grid()
            axarr[2,0].set_xlim([0,180])
            axarr[2,0].set_xlabel(r'$\theta$')
            axarr[2,0].set_xticks([0,30,60,90,120,150,180])
                    
            if (np.max(P11[:]) > 0.) :axarr[2,1].plot(ang, P44/P11)
            axarr[2,1].set_title(r'$P_{44}/P_{11}$')
            axarr[2,1].grid()
            axarr[2,1].set_xlim([0,180])
            axarr[2,1].set_xlabel(r'$\theta$')
            axarr[2,1].set_xticks([0,30,60,90,120,150,180])
    
    setp([a.get_xticklabels() for a in axarr[0, :]], visible=False)
    axarr[0,0].legend(loc='upper center',fontsize = 'medium',labelspacing=0.01)

    return fig, axarr

    
def profile_view(mlut, fig=None, ax=None, iw=0, kind='atm', zmax=None):
    '''
    visualization of a smartg MLUT profile from output

    Options:
        fig : fig object to be created or included in
        axarr : system of axes (2,2) to be created or used
        iw : in case of multi wavelength simulation, index of wavelength to be plotted
        kind : atmopsheric 'atm' or oceanic 'oc' profile
        zmax: max altitude or depth of the plot
    '''

    if (ax is None):
        fig, ax = subplots(1, 1)
        fig.set_size_inches(5, 5)
    
    nd = mlut['OD_'+kind].ndim
    if nd>1:
        wi = mlut['OD_'+kind].names.index('wavelength') # Wavelength index
        key = [slice(None)]*nd
        key[wi] = iw
        key=tuple(key)
        labw=r' at $%.1f nm$'%mlut.axes['wavelength'][iw]
        
    else:
        key=tuple([slice(None)])
        labw=''

    z = mlut.axis('z_'+kind, aslut=True)
    if kind=='oc': 
        sign=-1.
        func = diff2
    else:
        sign=1.    
        func = diff1
    Dz = z.apply(func)
    Dz = Dz.apply(abs,'Dz')
    Dtau     = sign * mlut['OD_'+kind    ].sub().__getitem__(key).apply(func,'Dtau')
    Dtau_Sca = sign * mlut['OD_sca_'+kind].sub().__getitem__(key).apply(func,'Dtau_Sca')
    Dtau_Abs = sign * mlut['OD_abs_'+kind].sub().__getitem__(key).apply(func,'Dtau_Abs')
    if kind=='atm':
        Dtau_ExtA = sign * mlut['OD_p'].sub().__getitem__(key).apply(func,'Dtau_ExtA')
        Dtau_ScaR = sign * mlut['OD_r'].sub().__getitem__(key).apply(func,'Dtau_ScaR')
        Dtau_AbsG = sign * mlut['OD_g'].sub().__getitem__(key).apply(func,'Dtau_AbsG')
        ssa_p = mlut['ssa_p_'+kind].sub().__getitem__(key)
        Dtau_ScaA = Dtau_ExtA * ssa_p
        Dtau_AbsA = Dtau_ExtA * (1. - ssa_p)
        if (np.max(Dtau_AbsA[:]) > 0.) : ax.semilogx((Dtau_AbsA/Dz)[:], z[:], 'r--',label=r'$\sigma_{abs}^{a+c}$')
        if (np.max(Dtau_ScaA[:]) > 0.) : ax.semilogx((Dtau_ScaA/Dz)[:], z[:], 'r',  label=r'$\sigma_{sca}^{a+c}$')
        if (np.max(Dtau_AbsG[:]) > 0.) : ax.semilogx((Dtau_AbsG/Dz)[:], z[:], 'g--',  label=r'$\sigma_{abs}^{gas}$')
        ax.semilogx((Dtau_ScaR/Dz)[:], z[:], 'b', label=r'$\sigma_{sca}^{R}$' )
        ax.set_xlim(1e-6,10)
        xlabel('Vertical profile'+labw + r' $(km^{-1})$')
        ylabel(r'$z (km)$')
        if zmax is None : zmax = max(100.,z.data.max())
        ax.set_ylim(0,zmax)
    else :
        Dtau_ExtP = sign * mlut['OD_p_oc'].sub().__getitem__(key).apply(func,'Dtau_ExtP')
        Dtau_ExtW = sign * mlut['OD_w'].sub().__getitem__(key).apply(func,'Dtau_ExtW')
        Dtau_AbsY = sign * mlut['OD_y'].sub().__getitem__(key).apply(func,'Dtau_AbsY')
        ssa_p = mlut['ssa_p_'+kind].sub().__getitem__(key)
        ssa_w = mlut['ssa_w'].sub().__getitem__(key)
        pine  = mlut['pine_oc'].sub().__getitem__(key)
        Dtau_ScaP = Dtau_ExtP * ssa_p
        Dtau_AbsP = Dtau_ExtP * (1. - ssa_p)
        Dtau_ScaW = Dtau_ExtW * ssa_w
        Dtau_AbsW = Dtau_ExtW * (1. - ssa_w)
        Dtau_Ine  = Dtau_Sca  * pine
        if (np.max(Dtau_AbsP[:]) > 0.) : ax.semilogx((Dtau_AbsP/Dz)[:], z[:], 'r--',label=r'$\sigma_{abs}^{p}$')
        if (np.max(Dtau_ScaP[:]) > 0.) : ax.semilogx((Dtau_ScaP/Dz)[:], z[:], 'r',  label=r'$\sigma_{sca}^{p}$')
        if (np.max(Dtau_AbsW[:]) > 0.) : ax.semilogx((Dtau_AbsW/Dz)[:], z[:], 'b--',label=r'$\sigma_{abs}^{w}$')
        if (np.max(Dtau_ScaW[:]) > 0.) : ax.semilogx((Dtau_ScaW/Dz)[:], z[:], 'b',  label=r'$\sigma_{sca}^{w}$')
        if (np.max(Dtau_AbsY[:]) > 0.) : ax.semilogx((Dtau_AbsY/Dz)[:], z[:], 'y--',label=r'$\sigma_{abs}^{y}$')
        if (np.max(Dtau_Ine[:])  > 0.) : ax.semilogx((Dtau_Ine/Dz)[:] , z[:], 'm:' ,label=r'$\sigma_{ine}^{}$')
        ax.set_xlim(1e-4,10)
        xlabel('Vertical profile'+labw + r' $(m^{-1})$')
        ylabel(r'$z (m)$')
        if zmax is None : zmax = min(-100.,z.data.min())
        ax.set_ylim(zmax,0)
    ax.semilogx((Dtau/Dz)[:], z[:], 'k.-', label=r'$\sigma_{ext}^{tot}$')
    ax.semilogx((Dtau_Abs/Dz)[:], z[:], 'k.--', label=r'$\sigma_{abs}^{tot}$')
    #ax.set_title('Vertical profile'+labw)
    ax.grid()
    ax.legend()

    try :
        ax2 = ax.twiny()
        nf= mlut['iphase_'+kind].__getitem__(key) 
        ax2.plot(nf[1:], mlut.axis('z_'+kind)[1:], 'm-', drawstyle='steps-post', label='i')
        ax2.set_xlabel('Phase Matrix index', color='m')
        ax2.tick_params('x', colors='m')
        ax2.xaxis.set_major_formatter(FormatStrFormatter('%i'))
        return fig, ax
    
    except:
        return fig, ax
    
    
def input_view(mlut, iw=0, kind='atm', zmax=None, ipha=None):
    ''' 
    visualization of a smartg MLUT profile and phase functions from output

    Options:
        iw : in case of multi wavelength simulation, index of wavelength to be plotted
        kind : atmopsheric 'atm' or oceanic 'oc' profile
        zmax: max altitude or depth of the plot
        ipha: absolute index of the phase function coming from Profile
    '''
    fig = figure()
    fig.set_size_inches(12,6)
    try:
        mlut['phase_'+kind]
        nstk = len(mlut['phase_'+kind][0,:,0])
        if nstk == 4:
            warnings.simplefilter('always', DeprecationWarning)
            warn_message = "\nGiving only 4 stk components is deprecated as of SMART-G 1.0.0 " + \
                           "and will removed in one of the next release.\n" + \
                           "Please provide phase with the 6 stk components in Iparper convention, where:\n" + \
                           "stk0 = F11, stk1=F12, stk2=F33, stk3=F34, stk4=F22 and stk5=F44."
            warnings.warn(warn_message, DeprecationWarning)
            ax1 = subplot2grid((2,3),(0,0))
            ax2 = subplot2grid((2,3),(0,1))
            ax3 = subplot2grid((2,3),(1,0))
            ax4 = subplot2grid((2,3),(1,1))
        
            axarr = np.array([[ax1,ax2],[ax3,ax4]])
        
            _,_= phase_view(mlut, iw=iw, axarr=axarr, kind=kind, ipha=ipha)
            
            ax5 = subplot2grid((2,3),(0,2),rowspan=2,colspan=1)
            
            _,_= profile_view(mlut, iw=iw, ax=ax5, kind=kind, zmax=zmax)
        else:
            fig.set_size_inches(12,9)
            ax1 = subplot2grid((3,3),(0,0))
            ax2 = subplot2grid((3,3),(0,1))
            ax3 = subplot2grid((3,3),(1,0))
            ax4 = subplot2grid((3,3),(1,1))
            ax5 = subplot2grid((3,3),(2,0))
            ax6 = subplot2grid((3,3),(2,1))
        
            axarr = np.array([[ax1,ax2],[ax3,ax4],[ax5,ax6]])
        
            _,_= phase_view(mlut, iw=iw, axarr=axarr, kind=kind, ipha=ipha)
            
            ax7 = subplot2grid((3,3),(0,2),rowspan=2,colspan=1)
            
            _,_= profile_view(mlut, iw=iw, ax=ax7, kind=kind, zmax=zmax)
        
        tight_layout()
        
    except:
        _,_= profile_view(mlut, iw=iw, kind=kind, zmax=zmax)

def compare(mlut, mref, field='up (TOA)',errb=False, logI=False, U_sign=1, same_U_convention=True, U_symetry=True,
                  Nparam=4, vmax=None, vmin=None, emax=None, ermax=None, same_azimuth_convention=True,
                  azimuth=[0.,90.], title='', SZA_MAX=89., zenith_title=r'$SZA (°)$', errref=None):
    '''
    compare the results of two smartg runs : mlut vs mref in two different azimuth planes
    outputs: a figure
    keywords:
        field : name of the output level to be compared 
        errb  : error bar visible for mlut : should have been run with the stdev option
        LogI  : plot Intensity in log scale
        U_sign: change sign for U
        same_U_convention: mlut and mref have the same convention for U
        U_symetry:  U changes sign convention for the  two halves of the plane
        Nparam: number of parameters :  by defaut 4 for I,Q,U, DoLP; 5 adds V:, 2 keeps only I and DoLP 
        vmin,vmax: min and max values for parameters: list of length Nparam
        emax: max absolute error scale : length Nparam
        ermax: max relative error scale (in percent) : length Nparam
        same_azimuth_convention: mlut and mref have the same azimuth convention
        azimuth: list of azimuths
        title: plot title
        SZA_MAX: SZA max for the comparison
        errref : eventually intensity absolute error on refence points
    '''

    from pylab import subplots
    if vmax is None : vmax=[0.1]*Nparam 
    if vmin is None : vmin=[-0.1]*Nparam 
    if emax is None : emax=[0.1]*Nparam
    if ermax is None : ermax=[0.1]*Nparam
    stokesT = ['I','Q','U','V']
    stokes=stokesT[:Nparam-1]
    signT = [1,1,U_sign*1,1,1] # sign convention for both mluts
    sign=signT[:Nparam-1]+[1]
    if same_U_convention: diffsignT = [1,1,1,1,1]    # sign convention difference
    else: diffsignT = [1,1,-1,1,1]
    diffsign=diffsignT[:Nparam-1]+[1]
    if U_symetry: symetryT=[1,1,1,1,1]
    else: symetryT=[1,1,-1,1,1]
    symetry=symetryT[:Nparam-1]+[1]
    fig,ax = subplots(3,Nparam, sharey=False,sharex=True,gridspec_kw=dict(hspace=0.2,wspace=0.3))
    fig.set_size_inches(Nparam*3,8)
    fig.set_dpi=600
    fig.suptitle(title)
    
    for i in range(Nparam):
        if i!=Nparam-1 :
            S = mlut[stokes[i] + '_' + field]
            if S.names.index('Azimuth angles') == 0: th = S.axes[1]   
            else: th = S.axes[0]
            Sref = mref[stokes[i] + '_' + field]
            S.desc = mdesc(S.desc)
            if errb : E = mlut[stokes[i] + '_' + 'stdev' + '_' + field]
            if logI and stokes[i]=='I':
                S=S.apply(np.log10)
                Sref=Sref.apply(np.log10)
                S.desc=r'$log_{10}$ '+S.desc
        else:
            I=mlut['I' + '_' + field]
            Q=mlut['Q' + '_' + field]
            U=mlut['U' + '_' + field]
            Ip= ((Q*Q+U*U).apply(np.sqrt))
            S= (Ip/I) * 100
            S.desc= r'$DoLP' + I.desc[1:3] + I.desc[4:] + '$'
            S.desc= 'DoLP' + I.desc[1:]
            S.desc = mdesc(S.desc)
            
            if errb: 
                dI=mlut['I' + '_' + 'stdev' + '_' + field]
                dQ=mlut['Q' + '_' + 'stdev' + '_' + field]
                dU=mlut['U' + '_' + 'stdev' + '_' + field]
                dIp= ((dQ*dQ+dU*dU).apply(np.sqrt))
                E = (dI/I + dIp/Ip) * S
            Iref=mref['I' + '_' + field]
            Qref=mref['Q' + '_' + field]
            Uref=mref['U' + '_' + field]
            Sref= (((Qref*Qref+Uref*Uref).apply(np.sqrt))/Iref) * 100           
     
        vmi=vmin[i]
        vma=vmax[i]
        ema=emax[i]
        erma=ermax[i]

        for phi0,sym1,sym2,labref in [(azimuth[0],'r','-','ref.'),(azimuth[1],'g','-','')]:
        #for phi0,sym1,sym2,labref in [(azimuth[0],'r','.','ref.'),(azimuth[1],'g','.','')]:

            # both points at their own abscissas
            if same_azimuth_convention:

                if S.names.index('Azimuth angles') == 0: # check right order of axes
                    refp = sign[i]*Sref[Idx(phi0,round=True),:] # reference for >0 view angle
                    refm = sign[i]*Sref[Idx(180.-phi0,round=True),:] #      reference for <0 view angle
                    sp   = diffsign[i]*sign[i]*S[Idx(phi0),:]       #     simulation for >0 view angle
                    sm   = symetry[i]*diffsign[i]*sign[i]*S[Idx(180-phi0),:]
                    if errb:
                        dsp  = E[Idx(phi0),:]         #     simulation error for >0 view angle
                        dsm  = E[Idx(180.-phi0),:]
                    else:
                        (dsp,dsm) = (0,0)
                else:
                    refm = sign[i]*Sref.swapaxes(0,1)[Idx(phi0,round=True),:] # reference for >0 view angle
                    refp = sign[i]*Sref.swapaxes(0,1)[Idx(180.-phi0,round=True),:] #      reference for <0 view angle
                    sp   = diffsign[i]*sign[i]*S.swapaxes(0,1)[Idx(phi0),:]       #     simulation for >0 view angle
                    sm   = symetry[i]*diffsign[i]*sign[i]*S.swapaxes(0,1)[Idx(180-phi0),:]
                    if errb:
                        dsp  = E.swapaxes(0,1)[Idx(phi0),:]         #     simulation error for >0 view angle
                        dsm  = E.swapaxes(0,1)[Idx(180.-phi0),:]
                    else:
                        (dsp,dsm) = (0,0)
                        
            else:
                refp = sign[i]*Sref[Idx(180.-phi0,round=True),:] # reference for >0 view angle
                refm = sign[i]*Sref[Idx(phi0,round=True),:] #      reference for <0 view angle
                sp   = diffsign[i]*sign[i]*S[Idx(phi0),:]       #     simulation for >0 view angle
                sm   = symetry[i]*diffsign[i]*sign[i]*S[Idx(180-phi0),:]
                if errb:
                    dsp  = E[Idx(phi0),:]         #     simulation error for >0 view angle
                    dsm  = E[Idx(180-phi0),:]
                else:
                    (dsp,dsm) = (0,0)
                    
            ax[0,i].plot(th, refp,'k'+'.')
            ax[0,i].plot(-th,refm,'k'+'.',label=labref)
            ax[0,i].errorbar(th, sp, fmt=sym1+'')
            ax[0,i].errorbar(-th,sm, fmt=sym1+'', \
                        label=r'$\Phi=%.0f-%.0f$'%(phi0,180.-phi0))
            ax[0,i].set_ylim([vmi, vma])
            ax[0,i].set_xlim([-SZA_MAX, SZA_MAX])
            ax[0,i].ticklabel_format(axis='y', style='sci', scilimits=(-2,2))
            
            if logI and i==0:
                if errb:
                    ax[1,i].errorbar(th,10**sp-10**refp, yerr=dsp,\
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f-%.0f$'%(phi0,180.-phi0),ecolor='k',capsize=2)
                    ax[1,i].errorbar(-th,10**sm-10**refm,yerr=dsm,fmt=sym1+sym2,ecolor='k',capsize=2) 
                else:
                    ax[1,i].errorbar(th,10**sp-10**refp, \
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f-%.0f$'%(phi0,180.-phi0),ecolor='k',capsize=2)
                    ax[1,i].errorbar(-th,10**sm-10**refm,fmt=sym1+sym2,ecolor='k',capsize=2) 
    
            else:
                if errb:
                    ax[1,i].errorbar(th,sp-refp, yerr=dsp,\
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f-%.0f$'%(phi0,180.-phi0),ecolor=sym1,capsize=2)
                    ax[1,i].errorbar(-th,sm-refm,yerr=dsm,fmt=sym1+sym2,ecolor=sym1,capsize=2) 
                else:
                    ax[1,i].errorbar(th,sp-refp, \
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f-%.0f$'%(phi0,180.-phi0),ecolor=sym1,capsize=2)
                    ax[1,i].errorbar(-th,sm-refm,fmt=sym1+sym2,ecolor=sym1,capsize=2) 
            ax[1,i].set_ylim([-1*ema,ema])
            ax[1,i].set_xlim([-SZA_MAX,SZA_MAX])  

            if errb:
                ax[2,i].errorbar(th,(sp-refp)/refp*100, yerr=dsp/abs(refp)*100, \
                             fmt=sym1+sym2,label=r'$\Phi=%.0f-%.0f$'%(phi0,180.-phi0),ecolor=sym1,capsize=2)
                ax[2,i].errorbar(-th,(sm-refm)/refm*100, yerr= dsm/abs(refm)*100,fmt=sym1+sym2,ecolor=sym1,capsize=2)  
                if (i==0 and errref is not None):
                    ax[2,0].plot(th,errref/refp*100,sym1+'-.')
                    ax[2,0].plot(th,-errref/refp*100,sym1+'-.')
                    ax[2,0].plot(-th,errref/refm*100,sym1+'-.')
                    ax[2,0].plot(-th,-errref/refm*100,sym1+'-.')
            else:
                ax[2,i].errorbar(th,(sp-refp)/refp*100,\
                             fmt=sym1+sym2,label=r'$\Phi=%.0f-%.0f$'%(phi0,180.-phi0),ecolor='k',capsize=2)
                ax[2,i].errorbar(-th,(sm-refm)/refm*100,fmt=sym1+sym2,ecolor='k',capsize=2)  
            
            if i!=Nparam-1 : ax[2,i].set_ylim([-1*erma,erma])
            else : ax[2,i].set_ylim([-1*erma,erma])
                
            ax[2,i].set_xlim([-SZA_MAX, SZA_MAX])    
            ax[1,i].plot([-SZA_MAX,SZA_MAX],[0.,0.],'k--')
            ax[2,i].plot([-SZA_MAX,SZA_MAX],[0.,0.],'k--')
            ax[1,i].ticklabel_format(axis='y', style='sci', scilimits=(-2,2))

            ax[0,i].set_title(S.desc)   
            if i==0: 
 
                ax[0,i].legend(loc='upper center',fontsize = 8,labelspacing=0.0)
                #ax[1,i].text(-50.,ema*0.75,r'$N_{\Phi}$:%i, $N_{\theta}$:%i'%\
                #         (S.axes[0].shape[0],S.axes[1].shape[0]))
                ax[1,i].set_ylabel(r'$\Delta$')
                ax[2,i].set_ylabel(r'$\Delta (\%)$')
            ax[2,i].set_xlabel(zenith_title)
    return fig


def compare_spectrum(mlut, mref, field='up (TOA)',errb=False, logI=False, U_sign=1, same_U_convention=True, U_symetry=True,
                  Nparam=4, vmax=None, vmin=None, emax=None, ermax=None, same_azimuth_convention=True,
                  azimuth=[0.,90.], title='', LMIN=None, LMAX=None, lambda_title=r'$\lambda$ (nm)', errref=None):
    '''
    compare the results of two smartg runs : mlut vs mref in two different azimuth planes
    outputs: a figure
    keywords:
        field : name of the output level to be compared 
        errb  : error bar visible for mlut : should have been run with the stdev option
        LogI  : plot Intensity in log scale
        U_sign: change sign for U
        same_U_convention: mlut and mref have the same convention for U
        U_symetry:  U changes sign convention for the  two halves of the plane
        Nparam: number of parameters :  by defaut 4 for I,Q,U, DoLP; 5 adds V:, 2 keeps only I and DoLP 
        vmin,vmax: min and max values for parameters: list of length Nparam
        emax: max absolute error scale : length Nparam
        ermax: max relative error scale (in percent) : length Nparam
        same_azimuth_convention: mlut and mref have the same azimuth convention
        azimuth: list of azimuths
        title: plot title
        SZA_MAX: SZA max for the comparison
        errref : eventually intensity absolute error on refence points
    '''

    from pylab import subplots
    if vmax is None : vmax=[0.1]*Nparam 
    if vmin is None : vmin=[-0.1]*Nparam 
    if emax is None : emax=[0.1]*Nparam
    if ermax is None : ermax=[0.1]*Nparam
    stokesT = ['I','Q','U','V']
    stokes=stokesT[:Nparam-1]
    signT = [1,1,U_sign*1,1,1] # sign convention for both mluts
    sign=signT[:Nparam-1]+[1]
    if same_U_convention: diffsignT = [1,1,1,1,1]    # sign convention difference
    else: diffsignT = [1,1,-1,1,1]
    diffsign=diffsignT[:Nparam-1]+[1]
    if U_symetry: symetryT=[1,1,1,1,1]
    else: symetryT=[1,1,-1,1,1]
    symetry=symetryT[:Nparam-1]+[1]
    fig,ax = subplots(3,Nparam, sharey=False,sharex=True,gridspec_kw=dict(hspace=0.2,wspace=0.3))
    fig.set_size_inches(Nparam*3,8)
    fig.set_dpi=600
    fig.suptitle(title)
    
    for i in range(Nparam):
        if i!=Nparam-1 :
            S = mlut[stokes[i] + '_' + field]
            wav = S.axis('wavelength')   
            Sref = mref[stokes[i] + '_' + field]
            S.desc = mdesc(S.desc)
            if errb : E = mlut[stokes[i] + '_' + 'stdev' + '_' + field]
            if logI and stokes[i]=='I':
                S=S.apply(np.log10)
                Sref=Sref.apply(np.log10)
                S.desc=r'$log_{10}$ '+S.desc
        else:
            I=mlut['I' + '_' + field]
            Q=mlut['Q' + '_' + field]
            U=mlut['U' + '_' + field]
            Ip= ((Q*Q+U*U).apply(np.sqrt))
            S= (Ip/I) * 100
            S.desc= r'$DoLP' + I.desc[1:3] + I.desc[4:] + '$'
            S.desc= 'DoLP' + I.desc[1:]
            S.desc = mdesc(S.desc)
            
            if errb: 
                dI=mlut['I' + '_' + 'stdev' + '_' + field]
                dQ=mlut['Q' + '_' + 'stdev' + '_' + field]
                dU=mlut['U' + '_' + 'stdev' + '_' + field]
                dIp= ((dQ*dQ+dU*dU).apply(np.sqrt))
                E = (dI/I + dIp/Ip) * S
            Iref=mref['I' + '_' + field]
            Qref=mref['Q' + '_' + field]
            Uref=mref['U' + '_' + field]
            Sref= (((Qref*Qref+Uref*Uref).apply(np.sqrt))/Iref) * 100           
     
        if LMIN is None:
            LMIN=wav.min()
        if LMAX is None:
            LMAX=wav.max()
        vmi=vmin[i]
        vma=vmax[i]
        ema=emax[i]
        erma=ermax[i]

        sym1s = ['r', 'g','b','m']
        sym2s = ['-','-','-','-']
        labrefs=['ref.','','','']
        for k in range(len(azimuth)):
            phi0 = azimuth[k]
            sym1 = sym1s[k]
            sym2 = sym2s[k]
            labref = labrefs[k]

            # both points at their own abscissas
            if same_azimuth_convention:
                if S.names.index('Azimuth angles') == 0: # check right order of axes
                    refp = sign[i]*Sref[Idx(phi0,round=True),:] # reference for >0 view angle
                    sp   = diffsign[i]*sign[i]*S[Idx(phi0),:]       #     simulation for >0 view angle
                    if errb:
                        dsp  = E[Idx(phi0),:]         #     simulation error for >0 view angle
                    else:
                        dsp = 0
                else:
                    refp = sign[i]*Sref.swapaxes(0,1)[Idx(phi0,round=True),:] #      reference for <0 view angle
                    sp   = diffsign[i]*sign[i]*S.swapaxes(0,1)[Idx(phi0),:]       #     simulation for >0 view angle
                    if errb:
                        dsp  = E.swapaxes(0,1)[Idx(phi0),:]         #     simulation error for >0 view angle
                    else:
                        dsp = 0
                        
            else:
                refp = sign[i]*Sref[Idx(180.-phi0,round=True),:] # reference for >0 view angle
                sp   = diffsign[i]*sign[i]*S[Idx(phi0),:]       #     simulation for >0 view angle
                if errb:
                    dsp  = E[Idx(phi0),:]         #     simulation error for >0 view angle
                else:
                    dsp = 0
                    
            ax[0,i].plot(wav,refp,'k'+'.',label=labref)
            ax[0,i].errorbar(wav,sp, fmt=sym1+'', \
                        label=r'$\Phi=%.0f$'%(phi0))
            ax[0,i].set_ylim([vmi, vma])
            ax[0,i].set_xlim([LMIN,LMAX])  
            ax[0,i].ticklabel_format(axis='y', style='sci', scilimits=(-2,2))
            
            if logI and i==0:
                if errb:
                    ax[1,i].errorbar(wav,10**sp-10**refp, yerr=dsp,\
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f$'%(phi0),ecolor='k',capsize=2)
                else:
                    ax[1,i].errorbar(wav,10**sp-10**refp, \
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f$'%(phi0),ecolor='k',capsize=2)
    
            else:
                if errb:
                    ax[1,i].errorbar(wav,sp-refp, yerr=dsp,\
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f$'%(phi0),ecolor=sym1,capsize=2)
                else:
                    ax[1,i].errorbar(wav,sp-refp, \
                                 fmt=sym1+sym2,label=r'$\Phi=%.0f$'%(phi0),ecolor=sym1,capsize=2)
            ax[1,i].set_ylim([-1*ema,ema])
            ax[1,i].set_xlim([LMIN,LMAX])  

            if errb:
                ax[2,i].errorbar(wav,(sp-refp)/refp*100, yerr=dsp/abs(refp)*100, \
                             fmt=sym1+sym2,label=r'$\Phi=%.0f$'%(phi0),ecolor=sym1,capsize=2)
                if (i==0 and errref is not None):
                    ax[2,0].plot(wav,errref/refp*100,sym1+'-.')
                    ax[2,0].plot(wav,-errref/refp*100,sym1+'-.')
            else:
                ax[2,i].errorbar(wav,(sp-refp)/refp*100,\
                             fmt=sym1+sym2,label=r'$\Phi=%.0f$'%(phi0),ecolor='k',capsize=2)
            
            if i!=Nparam-1 : ax[2,i].set_ylim([-1*erma,erma])
            else : ax[2,i].set_ylim([-1*erma,erma])
                
            ax[2,i].set_xlim([LMIN, LMAX])    
            ax[1,i].plot([LMIN,LMAX],[0.,0.],'k--')
            ax[2,i].plot([LMIN,LMAX],[0.,0.],'k--')
            ax[1,i].ticklabel_format(axis='y', style='sci', scilimits=(-2,2))

            ax[0,i].set_title(S.desc)   
            if i==0: 
 
                ax[0,i].legend(loc='upper center',fontsize = 8,labelspacing=0.0)
                ax[1,i].set_ylabel(r'$\Delta$')
                ax[2,i].set_ylabel(r'$\Delta (\%)$')
            ax[2,i].set_xlabel(lambda_title)
    return fig


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import
from luts.luts import MLUT, LUT, Idx
from smartg.atmosphere import diff1
import numpy as np
from warnings import warn
from smartg.albedo import Albedo_cst
from smartg.config import NPSTK
from smartg.tools.phase import fournierForand, integ_phase, calc_iphase
from smartg.bandset import BandSet
from smartg.config import DIR_AUXDATA as dir_aux

def diff2(x):
    return np.ediff1d(x, to_end=[0.])


def read_aw(dir_aux):
    '''
    Read pure water absorption from pope&fry, 97 (<700nm)
    and palmer&williams, 74 (>700nm)
    '''

    # Pope&Fry
    with open(dir_aux / 'water' / 'pope97.dat', 'rb') as fp:
        for i in range(6): fp.readline()  # skip the first 6 lines
        data_pf = np.genfromtxt(fp)
    aw_pf = data_pf[:,1] * 100 #  convert from cm-1 to m-1
    lam_pf = data_pf[:,0]
    ok_pf = lam_pf <= 725

    # Palmer&Williams
    data_pw = np.genfromtxt(dir_aux / 'water' / 'palmer74.dat', skip_header=5)
    aw_pw = data_pw[::-1,1] * 100 #  convert from cm-1 to m-1
    lam_pw = data_pw[::-1,0]
    ok_pw = lam_pw > 725

    aw = LUT(
            np.array(list(aw_pf[ok_pf]) + list(aw_pw[ok_pw])),
            axes=[np.array(list(lam_pf[ok_pf]) + list(lam_pw[ok_pw]))]
            )

    return aw


class IOP_base(object):
    pass

class IOP(IOP_base):
    '''
    Profile of water properties defined by:
        * phase: phase functions
          LUT with dimensions [nwav, nz, stk, angle]
        * bp, bw: particle and water scattering coefficient in m-1
          numpy arrays, dimensions [nwav, nZ]
          by default, bw takes default values
        * atot, ap, aw: total, particle and water absorption coefficient in m-1
          numpy array, dimensions [nwav, nZ]
          either provide atot or ap
          if ap is provided, aw can be provided as well, otherwise it takes default values
        * Z: profile of depths in m
        * ALB: seafloor albedo (albedo object)

        NOTE: first item in dimension Z is not used
    '''
    def __init__(self, phase=None, bp=None, bw=None,
                 atot=None, ap=None, aw=None, aCDOM=None, Bp=None,
                 Z=[0, -10000], NANG=721, ang_trunc=5., pfwav=None, ALB=Albedo_cst(0.)):

        self.Z = np.array(Z, dtype='float')
        self.bp = bp
        self.bw = bw
        self.atot = atot
        self.ap = ap
        self.aw = aw
        self.phase = phase
        self.aCDOM = aCDOM
        self.Bp  = Bp
        self.ALB = ALB
        self.NANG= NANG
        self.ang_trunc= ang_trunc
        self.coef_trunc = None
        self.NLAYER = len(Z)
        self.pfwav = pfwav

        self.AW = read_aw(dir_aux)

    def calc(self, wav, use_old_calc_iphase=False):
        shp = [x.shape for x in [self.bp, self.bw, self.atot, self.ap, self.aw, self.aCDOM, self.Bp] if x is not None][0]

        if not isinstance(wav, BandSet):
            wav = BandSet(wav)
        wav = np.array(wav)

        # absorption
        if self.ap is None:
            ap = np.zeros(shp, dtype='float')
        else:
            ap = self.ap
        
        # multilayer case
        NZ = ap.size//wav.size
        wav2 = np.stack([wav]*NZ,axis=1)

        if self.aCDOM is None:
            aCDOM = np.zeros(shp, dtype='float')
        else:
            aCDOM = self.aCDOM

        if self.aw is None:
            aw = self.AW[Idx(wav2)]
        else:
            aw = self.aw

        if self.atot is None:
            atot = aw + ap + aCDOM
        else:
            atot = self.atot

        # scattering
        if self.bp is None:
            bp = np.zeros(shp, dtype='float')
        else:
            bp = self.bp

        if self.pfwav is None:
            self.pfwav = wav

        if (self.phase is None) and (self.Bp is None) and ((np.array(bp) > 0).any()):
            raise Exception('No phase function nor Bp has been provided, but bp>0')

        bw = self.bw
        if bw is None:
            bw = 19.3e-4*((wav2/550.)**-4.3)

        FQYC = np.zeros(shp, dtype='float')

        pro = MLUT()

        pro.add_axis('wavelength', wav[:])
        pro.add_axis('z_oc', self.Z)
        pro.add_dataset('T_oc', np.array([280.]*len(self.Z), dtype='float32'),
                        ['z_oc'])

        if (self.phase is None) and (self.Bp is not None) and ((np.array(bp) > 0).any()):
            self.phase, self.coef_trunc = self.calc_phase(self.pfwav[:])
            bp *= self.coef_trunc.data

        btot = bw + bp

        if self.phase is not None:

            pha = self.phase
            pha_, ipha = calc_iphase(pha, pro.axis('wavelength'), pro.axis('z_oc'), use_old_calc_iphase)

            pro.add_axis('theta_oc', pha.axis('theta_oc'))
            pro.add_dataset('phase_oc', pha_, ['iphase', 'stk', 'theta_oc'])
            pro.add_dataset('iphase_oc', ipha, ['wavelength', 'z_oc'])

        dz = - diff1(self.Z)
        #dz = - diff1(self.Z) * aw/aw
        tau_w   = - (aw   + bw  ) * dz
        tau_p   = - (ap + bp  ) * dz
        tau_y   = - (aCDOM             ) * dz
        tau_tot = - (atot + btot) * dz
        tau_sca = - (btot              ) * dz
        tau_abs = - (atot              ) * dz
        tau_ine = - (ap * FQYC) * dz

        with np.errstate(invalid='ignore'):
            ssa_w   = bw/(aw   + bw)
        ssa_w[np.isnan(ssa_w)] = 1.

        with np.errstate(invalid='ignore'):
            ssa_p   = bp/(ap + bp)
        ssa_p[np.isnan(ssa_p)] = 1.

        with np.errstate(invalid='ignore'):
            pmol    = bw/(bw   + bp)
        pmol[np.isnan(pmol)]   = 1.
        pmol[~np.isfinite(pmol)] = 1.

        with np.errstate(invalid='ignore'):
            pine = tau_ine/tau_sca
        pine[np.isnan(pine)] = 0.
        pine[~np.isfinite(pine)] = 0.

        with np.errstate(invalid='ignore'):
            ssa = tau_sca/tau_tot
        ssa[np.isnan(ssa)] = 1.

        pro.add_dataset('OD_w', np.cumsum(tau_w, out=tau_w, axis=1),
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated water optical thickness at each wavelength'})

        pro.add_dataset('OD_p_oc', np.cumsum(tau_p, out=tau_p, axis=1),
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated oceanic particles optical thickness at each wavelength'})

        pro.add_dataset('OD_y', np.cumsum(tau_y, out=tau_y, axis=1),
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated CDOM optical thickness at each wavelength'})

        pro.add_dataset('OD_oc', np.cumsum(tau_tot, out=tau_tot, axis=1),
                        ['wavelength', 'z_oc'])

        pro.add_dataset('OD_sca_oc', np.cumsum(tau_sca, out=tau_sca, axis=1),
                        ['wavelength', 'z_oc'])

        pro.add_dataset('OD_abs_oc', np.cumsum(tau_abs, out=tau_abs, axis=1),
                        ['wavelength', 'z_oc'])

        pro.add_dataset('pine_oc', pine,
                        ['wavelength', 'z_oc'])

        pro.add_dataset('pmol_oc', pmol,
                        ['wavelength', 'z_oc'])

        pro.add_dataset('ssa_oc', ssa,
                        ['wavelength', 'z_oc'])

        pro.add_dataset('ssa_p_oc', ssa_p,
                        ['wavelength', 'z_oc'])
        pro.add_dataset('ssa_w', ssa_w,
                        ['wavelength', 'z_oc'])
        # NEW !!!
        pro.add_dataset('FQY1_oc', FQYC,
                        ['wavelength', 'z_oc'])
        # NEW !!!

        pro.add_dataset('albedo_seafloor',
                        self.ALB.get(wav),
                        ['wavelength'])
        return pro



    def calc_phase(self, wav):
        '''
        Calculate the phase function and associated truncation factor
        as a MLUT
        '''
        nwav = len(wav)
        nz   = self.NLAYER

        # particles phase function
        # see Park & Ruddick, 05
        # https://odnature.naturalsciences.be/downloads/publications/park_appliedoptics_2005.pdf
        ang = np.linspace(0, np.pi, self.NANG, dtype='float64')    # angle in radians
        ff1 = fournierForand(ang, 1.117,3.695)[None,None,:]
        ff2 = fournierForand(ang, 1.05, 3.259)[None,None,:]

        itronc = int(self.NANG * self.ang_trunc/180.)
        pha = np.zeros((nwav, nz, NPSTK, self.NANG), dtype='float64')
        r1 = ((self.Bp - 0.002)/0.028)[:,:,None]

        pha[:,:,0,:] = 0.5*(r1*ff1 + (1-r1)*ff2)

        # truncate
        pha[:,:,0,:itronc] = pha[:,:,0,itronc][:,:,None]

        pha[:,:,1,:] = pha[:,:,0,:]
        pha[:,:,2,:] = 0.
        pha[:,:,3,:] = 0.

        pha[:,:,:,0] = 0.

        # normalize
        integ_ff = integ_phase(ang, (pha[:,:,0,:] + pha[:,:,1,:])/2.)
        pha *= 2./integ_ff[:,:,None,None]

        P = LUT(pha,  # stk, theta
            axes=[wav, self.Z, None, np.rad2deg(ang)],
            names=['wav_phase_oc', 'z_phase_oc', 'stk', 'theta_oc'],
           )
        coef_trunc = LUT(integ_ff[:,:], axes=[wav, self.Z], names=['wav_phase_oc', 'z_phase_oc'])

        return P, coef_trunc




class IOP_Rw(IOP_base):
    def __init__(self, ALB):
        '''
        Defines a model of water reflectance (lambertian under the surface)

        ALB: albedo object of the lambertian reflector
        '''
        self.ALB = ALB

    def calc(self, wav):
        '''
        Profile and phase function calculation at bands wav (nm)
        '''
        if not isinstance(wav, BandSet):
            wav = BandSet(wav)
        wav = np.array(wav)

        pro = MLUT()
        pro.add_axis('wavelength', wav[:])
        pro.add_axis('z_oc', np.zeros(2))
        shp = (len(wav), 2)

        pro.add_dataset('T_oc', np.array([280., 280.], dtype='float32'),
                        ['z_oc'])
        pro.add_dataset('OD_oc', np.zeros(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('OD_w', np.zeros(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('OD_p_oc', np.zeros(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('OD_sca_oc', np.zeros(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('OD_abs_oc', np.zeros(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('OD_y', np.zeros(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('pmol_oc', np.ones(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('pine_oc', np.ones(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('FQY1_oc', np.ones(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('ssa_oc', np.ones(shp, dtype='float32'),
                        ['wavelength', 'z_oc'])
        pro.add_dataset('albedo_seafloor',
                        self.ALB.get(wav), ['wavelength'])

        return pro


class IOP_1(IOP_base):
    '''
    IOP model
    using similar IOP parameterization as Polymer's PR model

    Parameters:
        chl: chlorophyll concentration in mg/m3
        NANG: number of angles for the phase function
        ang_trunc: truncation angle in degrees
        ALB: albedo of the sea floor (Albedo instance)
        DEPTH: depth in m
        pfwav: list of wavelengths at which the phase functions are calculated
        FQYC: Chorophyll a fluorescence Quantum Yield
    '''
    def __init__(self, chl, pfwav=None, ALB=Albedo_cst(0.),
                 DEPTH=10000, NANG=72001, ang_trunc=5., FQYC=0.0):
        self.chl = chl
        self.depth = float(DEPTH)
        self.NANG = NANG
        self.ang_trunc = ang_trunc
        self.ALB = ALB
        self.FQYC = FQYC
        if pfwav is None:
            self.pfwav = None
        else:
            self.pfwav = np.array(pfwav)

        #
        # read pure water absorption coefficient
        #
        self.aw = read_aw(dir_aux)

        # Bricaud (98)
        ap_bricaud = np.genfromtxt(dir_aux / 'water' / 'aph_bricaud_1998.txt',
                                   delimiter=',', skip_header=12)  # header is lambda,Ap,Ep,Aphi,Ephi
        self.BRICAUD = MLUT()
        self.BRICAUD.add_axis('wav', ap_bricaud[:,0])
        self.BRICAUD.add_dataset('A', ap_bricaud[:,1], axnames=['wav'])
        self.BRICAUD.add_dataset('E', 1-ap_bricaud[:,2], axnames=['wav'])


    def calc_iop(self, wav, coef_trunc=2.):
        '''
        inherent optical properties calculation

        wav: wavelength in nm
        coef_trunc: integrate (normalized to 2) of the truncated phase function
        -> scattering coefficient has to be multiplied by coef_trunc/2.
        '''
        if not isinstance(wav, BandSet):
            wav = BandSet(wav)
        wav = np.array(wav)


        chl = self.chl

        # pure water absorption
        aw = self.aw[Idx(wav[:], fill_value='extrema')]

        # phytoplankton absorption
        aphy = (self.BRICAUD['A'][Idx(wav, fill_value='extrema')]
                * (chl**self.BRICAUD['E'][Idx(wav, fill_value='extrema')]))


        # NEW !!!
        # chlorophyll fluorescence (scattering coefficient)
        FQYC = np.full_like(aphy, self.FQYC) # Fluorescence Quantum Yield for Chlorophyll
        FQYC[wav<370.]=0.
        FQYC[wav>690.]=0.
        # NEW !!!

        # CDM absorption central value
        # from Bricaud et al GBC, 2012 (data from nov 2007)
        fa = 1.
        aCDM443 = fa * 0.069 * (chl**1.070)

        S = 0.00262*(aCDM443**(-0.448))
        if (S > 0.025): S=0.025
        if (S < 0.011): S=0.011

        aCDM = aCDM443 * np.exp(-S*(wav - 443))

        # NEW !!!
        atot = aw + aphy * (1.-FQYC) + aCDM
        # NEW !!!

        # Pure Sea water scattering coefficient
        bw = 19.3e-4*((wav/550.)**-4.3)

        bp = 0.416*(chl**0.766)*550./wav

        bp *= coef_trunc/2.

        # NEW !!!
        btot = bw + bp + aphy * FQYC
        # NEW !!!

        #
        # backscattering coefficient
        #
        if chl < 2:
            v = 0.5*(np.log10(chl) - 0.3)
        else:
            v = 0
        Bp = 0.002 + 0.01*( 0.5-0.25*np.log10(chl))*((wav/550.)**v)

        return {'atot': atot,
                'aphy': aphy,
                'aCDM': aCDM,
                'aw': aw,
                'bw': bw,
                'btot': btot,
                'bp': bp,
                'Bp': Bp,
                'FQYC': FQYC
                }


    def calc(self, wav, phase=True, use_old_calc_iphase=False):
        '''
        Profile and phase function calculation

        wav: wavelength in nm
        '''
        if not isinstance(wav, BandSet):
            wav = BandSet(wav)

        wav = np.array(wav)
        pro = MLUT()
        pro.add_axis('wavelength', wav[:])
        pro.add_axis('z_oc', np.array([0., -self.depth]))
        pro.add_dataset('T_oc', np.array([280., 280.], dtype='float32'), ['z_oc'])

        if phase:
            if self.pfwav is None:
                wav_pha = wav
            else:
                wav_pha = self.pfwav
            Bp = self.calc_iop(wav_pha)['Bp']
            pha = self.phase(wav_pha, Bp)

            pha_, ipha = calc_iphase(pha['phase'], pro.axis('wavelength'), pro.axis('z_oc'), use_old_calc_iphase)

            # index with ipha and reshape to broadcast to [wav, z]
            coef_trunc = pha['coef_trunc'].data.ravel()[ipha][:,0]  # discard dimension 'z_oc'

            pro.add_axis('theta_oc', pha.axis('theta_oc'))
            pro.add_dataset('phase_oc', pha_, ['iphase', 'stk', 'theta_oc'])
            pro.add_dataset('iphase_oc', ipha, ['wavelength', 'z_oc'])
        else:
            coef_trunc = 2.

        iop = self.calc_iop(wav, coef_trunc=coef_trunc)

        shp = (len(wav), 2)

        tau_w = np.zeros(shp, dtype='float32')
        tau_y = np.zeros(shp, dtype='float32')
        tau_p = np.zeros(shp, dtype='float32')
        ssa_w = np.zeros(shp, dtype='float32')
        ssa_p = np.zeros(shp, dtype='float32')
        tau_y[:,1]   = - iop['aCDM'] * self.depth
        tau_w[:,1]   = -(iop['aw'] + iop['bw'])*self.depth
        tau_p[:,1]   = -(iop['aphy'] + iop['bp'])*self.depth
        ssa_w[:,1]   =  iop['bw']/(iop['aw'] + iop['bw'])
        with np.errstate(invalid='ignore'):
            ssa_p[:,1] = iop['bp']/(iop['aphy'] + iop['bp'])
        ssa_p[np.isnan(ssa_p)] = 0.

        pro.add_dataset('OD_w', tau_w,
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated water optical thickness at each wavelength'})

        pro.add_dataset('OD_p_oc', tau_p,
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated oceanic particles optical thickness at each wavelength'})

        pro.add_dataset('OD_y', tau_y,
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated CDOM optical thickness at each wavelength'})

        tau_tot = np.zeros(shp, dtype='float32')
        tau_tot[:,1] = - ((iop['atot'] + iop['btot']) * self.depth)
        pro.add_dataset('OD_oc', tau_tot,
                        ['wavelength', 'z_oc'])

        tau_sca = np.zeros(shp, dtype='float32')
        tau_sca[:,1] = - (iop['btot'] * self.depth)
        pro.add_dataset('OD_sca_oc', tau_sca,
                        ['wavelength', 'z_oc'])

        tau_abs = np.zeros(shp, dtype='float32')
        tau_abs[:,1] = - (iop['atot'] * self.depth)
        pro.add_dataset('OD_abs_oc', tau_abs,
                        ['wavelength', 'z_oc'])

        pmol = np.ones(shp, dtype='float32')
        pmol[:,1] = iop['bw']/(iop['bw']+iop['bp'])
        pro.add_dataset('pmol_oc', pmol,
                        ['wavelength', 'z_oc'])

        with np.errstate(invalid='ignore'):
            ssa = tau_sca/tau_tot
        ssa[np.isnan(ssa)] = 1.

        pro.add_dataset('ssa_oc', ssa,
                        ['wavelength', 'z_oc'])
        pro.add_dataset('ssa_p_oc', ssa_p,
                        ['wavelength', 'z_oc'])
        pro.add_dataset('ssa_w', ssa_w,
                        ['wavelength', 'z_oc'])

        # NEW !!!
        tau_ine = np.ones(shp, dtype='float32')
        tau_ine[:,1] = -((iop['aphy'] * iop['FQYC']) * self.depth)
        with np.errstate(invalid='ignore'):
            pine = tau_ine/tau_sca
        pine[np.isnan(pine)] = 0.
        pro.add_dataset('pine_oc', pine,
                        ['wavelength', 'z_oc'])

        FQY1 = np.zeros(shp, dtype='float32')
        FQY1[:,1] = iop['FQYC']
        pro.add_dataset('FQY1_oc', FQY1,
                        ['wavelength', 'z_oc'])

        # NEW !!!
        pro.add_dataset('albedo_seafloor',
                        self.ALB.get(wav),
                        ['wavelength'])


        return pro

    def phase(self, wav, Bp):
        '''
        Calculate the phase function and associated truncation factor
        as a MLUT
        Bp is the backscattering ratio
        '''
        if not isinstance(wav, BandSet):
            wav = BandSet(wav)
        wav = np.array(wav)

        nwav = len(wav)

        # particles phase function
        # see Park & Ruddick, 05
        # https://odnature.naturalsciences.be/downloads/publications/park_appliedoptics_2005.pdf
        ang = np.linspace(0, np.pi, self.NANG, dtype='float64')    # angle in radians
        ff1 = fournierForand(ang, 1.117,3.695)[None,:]
        ff2 = fournierForand(ang, 1.05, 3.259)[None,:]

        itronc = int(self.NANG * self.ang_trunc/180.)
        pha = np.zeros((nwav, 1, NPSTK, self.NANG), dtype='float64')
        r1 = ((Bp - 0.002)/0.028)[:,None]

        pha[:,0,0,:] = 0.5*(r1*ff1 + (1-r1)*ff2)

        # truncate
        pha[:,0,0,:itronc] = pha[:,0,0,itronc][:,None]

        pha[:,0,1,:] = pha[:,0,0,:]
        pha[:,0,2,:] = 0.
        pha[:,0,3,:] = 0.

        pha[:,:,:,0] = 0.

        # normalize
        integ_ff = integ_phase(ang, (pha[:,0,0,:] + pha[:,0,1,:])/2.)
        pha *= 2./integ_ff[:,None,None,None]

        # create output MLUT
        result = MLUT()
        result.add_axis('wav_phase_oc', wav)
        result.add_axis('z_phase_oc', np.array([0.]))
        result.add_axis('theta_oc', ang*180./np.pi)
        result.add_dataset('phase', pha, ['wav_phase_oc', 'z_phase_oc', 'stk', 'theta_oc'])
        result.add_dataset('coef_trunc', integ_ff[:,None], ['wav_phase_oc', 'z_phase_oc'])

        return result


class IOP_profile(IOP_base):
    '''
    IOP model described in Zhai et al, optics express, 2017

    Parameters:
        chl: chlorophyll concentration in mg/m3 at the surface
        NANG: number of angles for the phase function
        ang_trunc: truncation angle in degrees
        ALB: albedo of the sea floor (Albedo instance)
        DEPTH: depth in m
        pfwav: list of wavelengths at which the phase functions are calculated
        NLAYER: Number of vertical layers
        Zeu   : Euphotic Depth (m), default is computed from climatology
        MIXED: Mixed or Statified waters
        FQYC : Fluorescence Quantum Yield for Chlorophyll
    '''
    def __init__(self, chls, pfwav=None, ALB=Albedo_cst(0.),
                 DEPTH=300., NANG=7201, ang_trunc=5., NLAYER=20, Zeu=None, MIXED=False, FQYC=0.):
        self.chls = chls
        self.depth = float(DEPTH)
        self.NANG = NANG
        self.ang_trunc = ang_trunc
        self.ALB = ALB
        self.FQYC = FQYC
        self.NLAYER=NLAYER
        if pfwav is None:
            self.pfwav = None
        else:
            self.pfwav = np.array(pfwav)

        #
        # read pure water absorption coefficient
        #
        self.aw = read_aw(dir_aux)

        # Bricaud (98)
        # Absorption of the phytoplankton
        ap_bricaud = np.genfromtxt(dir_aux / 'water' / 'aph_bricaud_1998.txt',
                                   delimiter=',', skip_header=12)  # header is lambda,Ap,Ep,Aphi,Ephi
        # Add extension to 360 nm (Wei et al., 2016)
        # spectral slope of aph is symetrical wrt 440 nm in the 360-520 spectral range
        wUV = np.linspace(360., 398., num=20)
        A   = ap_bricaud[:,1]
        B   = 1.-ap_bricaud[:,2]
        w   = ap_bricaud[:,0]
        ii  = np.where((w<=520.) & (w>480.))
        AUV = np.zeros_like(wUV)
        BUV = np.zeros_like(wUV)
        AUV[::-1] = A[ii]
        BUV[::-1] = B[ii]
        self.BRICAUD = MLUT()
        self.BRICAUD.add_axis('wav', np.concatenate((wUV,ap_bricaud[:,0])))
        self.BRICAUD.add_dataset('A', np.concatenate((AUV,A)), axnames=['wav'])
        self.BRICAUD.add_dataset('E', np.concatenate((BUV,B)), axnames=['wav'])

        ## Determine Chl vertical profile
        #1. Determine chlorophyll interagted column until Euphotic Depth Zeu 
        if Zeu is None:
            if not MIXED:
                if (chls > 1.): chl_zeu = 37.7*chls**0.615 
                else: chl_zeu = 36.1*chls**0.357
            else:
                chl_zeu = 42.1*chls**0.538
            Zeu = 568.2*chl_zeu**(-0.746)
         
        #2. Derive Euphotic Depth and make vertical grid
        Zmax= min(DEPTH, 5*Zeu)
        self.z = - np.linspace(0, Zmax, num=NLAYER+1)
        #self.z = np.linspace(0, Zmax, num=NLAYER+1)

        #3. Introduce reduced concentration chi and reduced depth zeta
        # Stratified Trophic case 1 parametrization
        #   see J. Uitz, H. Claustre, A. Morel, and S. B. Hooker, 
        #   “Vertical distribution of phytoplankton communities in open ocean: 
        #   An assessment based on surface chlorophyll,” J. Geophys. Res. 111, C08005 (2006).
        self.chi_b    = 0.471
        self.s        = 0.135
        self.chi_max  = 1.572
        self.zeta_max = 0.969
        self.Dzeta    = 0.393
        zeta = abs(self.z/Zeu)
        chl  = self.chls*self.chi(zeta)/self.chi(0.)
        chl[chl<0.]=1e-8
        self.chl = chl
        self.chlmean = np.trapz(chl,-self.z)/DEPTH
        #self.chlmean = np.trapz(chl,self.z)/DEPTH

    def chi(self, zeta):
        return self.chi_b - self.s*zeta + self.chi_max*np.exp(-((zeta-self.zeta_max)/self.Dzeta)**2)


    def calc_iop(self, wav, coef_trunc=2., p1=0.33, R1=0.5, R2=0.5):
        '''
        inherent optical properties calculation

        wav: wavelength in nm
        coef_trunc: integrate (normalized to 2) of the truncated phase function
        -> scattering coefficient has to be multiplied by coef_trunc/2.
        p1 and R1 are parameters explained in Zhai et al. 2017, related to particles extinction
        R2 is related to CDOM absorption
        '''
        if not isinstance(wav, BandSet):
            wav = BandSet(wav)
        wav = np.array(wav)

        chl = self.chl
        chl2, wav2 = np.meshgrid(chl,wav)

        # pure water absorption
        aw = self.aw[Idx(wav2[:])]
        # Pure Sea water scattering coefficient
        bw = 19.3e-4*((wav2/550.)**-4.32)

        # phytoplankton absorption
        #aphy = (self.BRICAUD['A'][Idx(wav2, fill_value='extrema')]
        #        * (chl2**self.BRICAUD['E'][Idx(wav2, fill_value='extrema')]))
        #aphy440B = (self.BRICAUD['A'][Idx(440., fill_value='extrema')]
        #        * (chl2**self.BRICAUD['E'][Idx(440., fill_value='extrema')]))
        # specific phytoplankton absorption
        chl2star=np.full_like(chl2, 1.)
        aphystar = (self.BRICAUD['A'][Idx(wav2, fill_value='extrema')]
                * (chl2star**self.BRICAUD['E'][Idx(wav2, fill_value='extrema')]))
        aphy = aphystar * chl2
        aphystar440 = (self.BRICAUD['A'][Idx(440., fill_value='extrema')]
                * (chl2star**self.BRICAUD['E'][Idx(wav2, fill_value='extrema')]))
        aphy440 = aphystar440 * chl2
        # phytoplankton covariant particles extinction
        ## Zhai et al., 2017
        # cp550 = p1*chl2**0.57
        # n1    = -0.4 + (1.6+1.2*R1)/(1+chl2**0.5)
        # cp    = cp550*(550./wav2)**n1
        # bp    = cp-aphy

        ## return to IOP_1 definition
        #bp = 0.416*(chl2**0.766)*550./wav2
        piz440=0.68
        bp440 = aphy440 * piz440/(1-piz440)
        bp = bp440 *(wav2/440.)**(-1.)
        #bp = 0.416*(chl2**0.766)*(wav2/550.)**(-1.)*6
        # correct for phase function truncation
        bp *= coef_trunc/2.
        # NEW !!!
        # chlorophyll fluorescence (scattering coefficient)
        FQYC = np.full_like(aphy, self.FQYC) # Fluorescence Quantum Yield for Chlorophyll
        FQYC[wav2<370.]=0.
        FQYC[wav2>690.]=0.

        # CDOM covariant absorption
        ## Zhai et al., 2017
        # p2       = 0.3 + (5.7*R2*aphy440)/(0.02+aphy440)
        # aCDOM440 = p2 * aphy440
        # aCDOM    = aCDOM440 *np.exp(-0.014*(wav2-440.))
        # CDM absorption central value

        ## return to IOP_1 definition
        # from Bricaud et al GBC, 2012 (data from nov 2007)
        #fa = 1.
        #aCDM443 = fa * 0.069 * (chl2**1.070)
        aCDM440 = 0.24*aphy440**0.43

        #S = 0.00262*(aCDM443**(-0.448))
        #S[S > 0.025] = 0.025
        #S[S < 0.011] = 0.011
        S=0.02

        aCDOM = aCDM440 * np.exp(-S*(wav2 - 440))
        #aCDOM = aCDM443 * np.exp(-S*(wav2 - 443))

        # NEW !!!
        atot = aw + aphy*(1.-FQYC) + aCDOM
        # NEW !!!

        # NEW !!!
        SPM = 0. # g/m3
        #SPM = 10. # g/m3
        gamma=0.5
        bbpnap650 = 10**(1.03*np.log10(SPM) - 2.06) # Neukermans et al 2012
        bbpnap = bbpnap650*(wav2/650.)**(-gamma)
        Bpnap = np.zeros_like(aphy)
        Bpnap[:] = 0.04
        bpnap  = bbpnap/Bpnap
        bp += bpnap
        btot = bw + bp + aphy* FQYC
        # NEW !!!

        #
        # backscattering coefficient
        #
        v = np.zeros_like(aphy)
        good=np.where(chl2<2.)
        v[good] = 0.5*(np.log10(chl2[good]) - 0.3)
        #Bp = 0.002 + 0.01*( 0.5-0.25*np.log10(chl2))*((wav2/550.)**v)
        #Bp[np.isinf(Bp)]=0.5
        #Bp[np.isnan(Bp)]=0.5


        Bp = Bpnap

        return {'atot': atot,
                'aphy': aphy,
                'aCDOM':aCDOM,
                'aw': aw,
                'bw': bw,
                'btot': btot,
                'bp': bp,
                'Bp': Bp,
                'FQYC': FQYC
                }

    def calc(self, wav, phase=True, use_old_calc_iphase=False):
        '''
        Profile and phase function calculation

        wav: wavelength in nm
        '''
        if not isinstance(wav, BandSet):
            wav = BandSet(wav)

        wav = np.array(wav)
        pro = MLUT()
        pro.add_axis('wavelength', wav[:])
        pro.add_axis('z_oc', self.z)
        #pro.add_axis('z_oc', -self.z)
        pro.add_dataset('T_oc', np.array([280.]*len(self.z), dtype='float32'), ['z_oc'])


        if phase:
            if self.pfwav is None:
                wav_pha = wav
            else:
                wav_pha = self.pfwav
            Bp = self.calc_iop(wav_pha)['Bp']
            pha = self.phase(wav_pha, Bp[:,1:])

            pha_, ipha = calc_iphase(pha['phase'], pro.axis('wavelength'), pro.axis('z_oc'), use_old_calc_iphase)

            # index with ipha and reshape to broadcast to [wav, z]
            coef_trunc = pha['coef_trunc'].data.ravel()[ipha][:,:]

            pro.add_axis('theta_oc', pha.axis('theta_oc'))
            pro.add_dataset('phase_oc', pha_, ['iphase', 'stk', 'theta_oc'])
            pro.add_dataset('iphase_oc', ipha, ['wavelength', 'z_oc'])
        else:
            coef_trunc = 2.

        iop = self.calc_iop(wav, coef_trunc=coef_trunc)

        dz   = -diff1(self.z)
        #dz   = diff1(self.z)
        zeros = np.zeros((len(wav),1))
        for key in iop.keys():
            iop[key] = np.append(zeros, iop[key][:,:-1], axis=1)

        tau_w   = - (iop['aw']   + iop['bw']  ) * dz
        tau_p   = - (iop['aphy'] + iop['bp']  ) * dz
        tau_y   = - (iop['aCDOM']             ) * dz
        tau_tot = - (iop['atot'] + iop['btot']) * dz
        tau_sca = - (iop['btot']              ) * dz
        tau_abs = - (iop['atot']              ) * dz
        tau_ine = - (iop['aphy'] * iop['FQYC']) * dz


        with np.errstate(invalid='ignore'):
            ssa_w   = iop['bw']/(iop['aw']   + iop['bw'])
        ssa_w[np.isnan(ssa_w)] = 1.

        with np.errstate(invalid='ignore'):
            ssa_p   = iop['bp']/(iop['aphy'] + iop['bp'])
        ssa_p[np.isnan(ssa_p)] = 1.

        with np.errstate(invalid='ignore'):
            pmol    = iop['bw']/(iop['bw']   + iop['bp'])
        pmol[np.isnan(pmol)]   = 1.

        with np.errstate(invalid='ignore'):
            pine = tau_ine/tau_sca
        pine[np.isnan(pine)] = 0.

        with np.errstate(invalid='ignore'):
            ssa = tau_sca/tau_tot
        ssa[np.isnan(ssa)] = 1.

        pro.add_dataset('OD_w', np.cumsum(tau_w, out=tau_w, axis=1),
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated water optical thickness at each wavelength'})

        pro.add_dataset('OD_p_oc', np.cumsum(tau_p, out=tau_p, axis=1),
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated oceanic particles optical thickness at each wavelength'})

        pro.add_dataset('OD_y', np.cumsum(tau_y, out=tau_y, axis=1),
                        ['wavelength', 'z_oc'],
                        attrs={'description':
                               'Cumulated CDOM optical thickness at each wavelength'})

        pro.add_dataset('OD_oc', np.cumsum(tau_tot, out=tau_tot, axis=1),
                        ['wavelength', 'z_oc'])

        pro.add_dataset('OD_sca_oc', np.cumsum(tau_sca, out=tau_sca, axis=1),
                        ['wavelength', 'z_oc'])

        pro.add_dataset('OD_abs_oc', np.cumsum(tau_abs, out=tau_abs, axis=1),
                        ['wavelength', 'z_oc'])

        pro.add_dataset('pine_oc', pine,
                        ['wavelength', 'z_oc'])

        pro.add_dataset('pmol_oc', pmol,
                        ['wavelength', 'z_oc'])

        pro.add_dataset('ssa_oc', ssa,
                        ['wavelength', 'z_oc'])

        pro.add_dataset('ssa_p_oc', ssa_p,
                        ['wavelength', 'z_oc'])
        pro.add_dataset('ssa_w', ssa_w,
                        ['wavelength', 'z_oc'])
        # NEW !!!
        pro.add_dataset('FQY1_oc', iop['FQYC'],
                        ['wavelength', 'z_oc'])
        # NEW !!!

        pro.add_dataset('albedo_seafloor',
                        self.ALB.get(wav),
                        ['wavelength'])

        return pro


    def phase(self, wav, Bp):
        '''
        Calculate the phase function and associated truncation factor
        as a MLUT
        Bp is the backscattering ratio
        '''
        nwav = len(wav)
        nz   = self.NLAYER

        # particles phase function
        # see Park & Ruddick, 05
        # https://odnature.naturalsciences.be/downloads/publications/park_appliedoptics_2005.pdf
        ang = np.linspace(0, np.pi, self.NANG, dtype='float64')    # angle in radians
        ff1 = fournierForand(ang, 1.117,3.695)[None,None,:]
        ff2 = fournierForand(ang, 1.05, 3.259)[None,None,:]

        itronc = int(self.NANG * self.ang_trunc/180.)
        pha = np.zeros((nwav, nz, NPSTK, self.NANG), dtype='float64')
        r1 = ((Bp - 0.002)/0.028)[:,:,None]

        pha[:,:,0,:] = 0.5*(r1*ff1 + (1-r1)*ff2)

        # truncate
        pha[:,:,0,:itronc] = pha[:,:,0,itronc][:,:,None]

        pha[:,:,1,:] = pha[:,:,0,:]
        pha[:,:,2,:] = 0.
        pha[:,:,3,:] = 0.

        pha[:,:,:,0] = 0.

        # normalize
        integ_ff = integ_phase(ang, (pha[:,:,0,:] + pha[:,:,1,:])/2.)
        pha *= 2./integ_ff[:,:,None,None]

        # create output MLUT
        result = MLUT()
        result.add_axis('wav_phase_oc', wav)
        result.add_axis('z_phase_oc', self.z[:-1])
        #result.add_axis('z_phase_oc', -self.z[:-1])
        result.add_axis('theta_oc', ang*180./np.pi)
        result.add_dataset('phase', pha, ['wav_phase_oc', 'z_phase_oc', 'stk', 'theta_oc'])
        result.add_dataset('coef_trunc', integ_ff[:,:], ['wav_phase_oc', 'z_phase_oc'])

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This module is deprecated. It will be removed in one of the next versions of smartg.
# All the classes/functions are now available in the geoclide package.

from smartg.geometry import Vector, Point, Normal, Ray
from smartg.geometry import Dot, Cross, Normalize, CoordinateSystem, \
    Distance, FaceForward
from warnings import warn


#####################################################################################
class DifferentialGeometry(object):
    '''
    Declare common parameters, enable to shade some parameters and then declare in
    other files several shapes without the need to distinguish them.
    '''
    def __init__(self, p, dpdu, dpdv, uu, vv, shape = None):
        warn_message = 'DifferentialGeometry is depracated as of smartg v1.1.0. ' + \
                       'and will be removed in one of the next versions of smartg.'
        warn(warn_message, DeprecationWarning)
        if isinstance(p, Point) and isinstance(dpdu, Vector) and \
           isinstance(dpdv, Vector):
            self.p = p
            self.dpdu = dpdu
            self.dpdv = dpdv
            # self.dndu = dndu
            # self.dndv = dndv
            self.nn = Normal(Normalize(Cross(self.dpdu, self.dpdv)))
            self.u = uu
            self.v = vv
            # self.dudx = 0
            # self.dvdx = 0
            # self.dudy = 0
            # self.dvdy = 0
            self.shape = shape
        else:
            raise NameError('Problem with the type of argument(s)')
#####################################################################################

if __name__ == '__main__':
    
    a = Vector(10, 2, 3)
    b = Point(2, 4, 6)
    c = Vector(b)*2
    p1 = Point(4, 8, 12)
    print("a =", a)
    print("b =", b)
    print("c =", c)
    print("Dot(a, c) =", Dot(a, c))
    print("Dot(a, b) =", Dot(a, Vector(b)))
    print("Cross(a, c) =", Cross(a, c))
    print("Lengh(a) =", a.Lengh())
    print("normalize(a) =", Normalize(a))
    print("Lengh of Normalize(a) =", Normalize(a).Lengh())

    v1 = Vector(2, 1, 0)
    v2, v3 = CoordinateSystem(v1)

    print("v1 =", v1)
    print("v2 =", v2)
    print("v3 =", v3)
    print("p2 =", p1-b)
    print("type(p2) =", type(p1-b))
    print("Distance(b, p1) =",  Distance(b, p1))
    print("a*2 =", a*2)

    vec1 = Vector (-2, 4, -1)
    vec2 = Vector (5, 1, 3)

    print("FaceForward(vec1, vec2) =", FaceForward(vec1, vec2))

    rayon = Ray(p1, vec2)
    print("p1 =", rayon.o)
    print("vec2 =", rayon.d)
    print("rayon(t=1) =", rayon[1])
    print("rayon(t=2) =", rayon[2])



#!/usr/bin/env python
# vim:fileencoding=utf-8


from numpy import pi, ndarray, float32, float64

def swscat(lam, delta=0.039, Tc=17., S=35.):
    '''
    Scattering coefficient of sea water

    Xiaodong Zhang, Lianbo Hu, and Ming-Xia He (2009), Scatteirng by pure
    seawater: Effect of salinity, Optics Express, Vol. 17, No. 7, 5698-5710

    lam (nm): wavelength
    Tc: temperauter in degree Celsius, must be a scalar
    S: salinity, must be scalar
    delta: depolarization ratio, if not provided, default = 0.039 will be
    used. (Farinato and Roswell (1976))
    betasw: volume scattering at angles defined by theta. Its size is [x y],
    where x is the number of angles (x = length(theta)) and y is the number
    of wavelengths in lambda (y = length(lambda))
    beta90sw: volume scattering at 90 degree. Its size is [1 y]
    bw: total scattering coefficient. Its size is [1 y]
    for backscattering coefficients, divide total scattering by 2

    adapted from matlab script betasw_ZHH2009.m
    '''
    
    assert isinstance(lam, (float, ndarray))
    if isinstance(lam, ndarray):
        assert (lam.dtype in [float32, float64])

    assert isinstance(Tc, float)
    assert isinstance(S, float)


    # values of the constants
    Na = 6.0221417930e23     #  Avogadro's constant
    Kbz = 1.3806503e-23      #  Boltzmann constant
    Tk = Tc+273.15           #  Absolute tempearture
    M0 = 18e-3               #  Molecular weigth of water in kg/mol

    # nsw: absolute refractive index of seawater
    # dnds: partial derivative of seawater refractive index w.r.t. salinity
    (nsw, dnds) = RInw(lam,Tc,S)

    # isothermal compressibility is from Lepple & Millero (1971,Deep
    # Sea-Research), pages 10-11
    # The error ~ +/-0.004e-6 bar^-1
    IsoComp = BetaT(Tc,S)

    # density of water and seawater,unit is Kg/m^3, from UNESCO,38,1981
    density_sw = rhou_sw(Tc, S)

    # water activity data of seawater is from Millero and Leung (1976,American
    # Journal of Science,276,1035-1077). Table 19 was reproduced using
    # Eq.(14,22,23,88,107) then were fitted to polynominal equation.
    # dlnawds is partial derivative of natural logarithm of water activity
    # w.r.t.salinity
    dlnawds = dlnasw_ds(Tc, S)

    # density derivative of refractive index from PMH model
    DFRI = PMH(nsw)   # PMH model

    # volume scattering at 90 degree due to the density fluctuation
    beta_df = pi*pi/2*((lam*1e-9)**(-4))*Kbz*Tk*IsoComp*DFRI**2*(6+6*delta)/(6-7*delta)
    # volume scattering at 90 degree due to the concentration fluctuation
    flu_con = S*M0*dnds**2/density_sw/(-dlnawds)/Na
    beta_cf = 2*pi*pi*((lam*1e-9)**(-4))*nsw**2*(flu_con)*(6+6*delta)/(6-7*delta)
    # total volume scattering at 90 degree
    beta90sw = beta_df+beta_cf
    bsw=8*pi/3*beta90sw*(2+delta)/(1+delta)

    # return bsw
    return bsw


def RInw(lam,Tc,S):
    # refractive index of air is from Ciddor (1996,Applied Optics)
    n_air = 1.0+(5792105.0/(238.0185-1/(lam/1e3)**2)+167917.0/(57.362-1/(lam/1e3)**2))/1e8

    # refractive index of seawater is from Quan and Fry (1994, Applied Optics)
    n0 = 1.31405; n1 = 1.779e-4 ; n2 = -1.05e-6 ; n3 = 1.6e-8 ; n4 = -2.02e-6
    n5 = 15.868; n6 = 0.01155;  n7 = -0.00423;  n8 = -4382 ; n9 = 1.1455e6

    nsw = n0+(n1+n2*Tc+n3*Tc**2)*S+n4*Tc**2+(n5+n6*S+n7*Tc)/lam+n8/lam**2+n9/lam**3  # pure seawater
    nsw = nsw*n_air
    dnswds = (n1+n2*Tc+n3*Tc**2+n6/lam)*n_air

    return (nsw, dnswds)

def BetaT(Tc, S):
    # pure water secant bulk Millero (1980, Deep-sea Research)
    kw = 19652.21+148.4206*Tc-2.327105*Tc**2+1.360477e-2*Tc**3-5.155288e-5*Tc**4
    Btw_cal = 1/kw

    # isothermal compressibility from Kell sound measurement in pure water
    # Btw = (50.88630+0.717582*Tc+0.7819867e-3*Tc**2+31.62214e-6*Tc**3-0.1323594e-6*Tc**4+0.634575e-9*Tc**5)/(1+21.65928e-3*Tc)*1e-6

    # seawater secant bulk
    a0 = 54.6746-0.603459*Tc+1.09987e-2*Tc**2-6.167e-5*Tc**3
    b0 = 7.944e-2+1.6483e-2*Tc-5.3009e-4*Tc**2

    Ks =kw + a0*S + b0*S**1.5

    # calculate seawater isothermal compressibility from the secant bulk
    IsoComp = 1/Ks*1e-5  # unit is pa


    return IsoComp

def rhou_sw(Tc, S):

    # density of water and seawater,unit is Kg/m^3, from UNESCO,38,1981
    a0 = 8.24493e-1;  a1 = -4.0899e-3; a2 = 7.6438e-5; a3 = -8.2467e-7; a4 = 5.3875e-9
    a5 = -5.72466e-3; a6 = 1.0227e-4;  a7 = -1.6546e-6; a8 = 4.8314e-4
    b0 = 999.842594; b1 = 6.793952e-2; b2 = -9.09529e-3; b3 = 1.001685e-4
    b4 = -1.120083e-6; b5 = 6.536332e-9

    # density for pure water 
    density_w = b0+b1*Tc+b2*Tc**2+b3*Tc**3+b4*Tc**4+b5*Tc**5
    # density for pure seawater
    density_sw = density_w +((a0+a1*Tc+a2*Tc**2+a3*Tc**3+a4*Tc**4)*S+(a5+a6*Tc+a7*Tc**2)*S**1.5+a8*S**2)

    return density_sw

def dlnasw_ds(Tc, S):

    # water activity data of seawater is from Millero and Leung (1976,American
    # Journal of Science,276,1035-1077). Table 19 was reproduced using
    # Eqs.(14,22,23,88,107) then were fitted to polynominal equation.
    # dlnawds is partial derivative of natural logarithm of water activity
    # w.r.t.salinity
    # lnaw = (-1.64555e-6-1.34779e-7*Tc+1.85392e-9*Tc**2-1.40702e-11*Tc**3)+......
    #            (-5.58651e-4+2.40452e-7*Tc-3.12165e-9*Tc**2+2.40808e-11*Tc**3)*S+......
    #            (1.79613e-5-9.9422e-8*Tc+2.08919e-9*Tc**2-1.39872e-11*Tc**3)*S**1.5+......
    #            (-2.31065e-6-1.37674e-9*Tc-1.93316e-11*Tc**2)*S**2

    dlnawds = (-5.58651e-4+2.40452e-7*Tc-3.12165e-9*Tc**2+2.40808e-11*Tc**3)+ \
               1.5*(1.79613e-5-9.9422e-8*Tc+2.08919e-9*Tc**2-1.39872e-11*Tc**3)*S**0.5+ \
               2*(-2.31065e-6-1.37674e-9*Tc-1.93316e-11*Tc**2)*S

    return dlnawds

# density derivative of refractive index from PMH model
def PMH(n_wat):
    n_wat2 = n_wat**2
    n_density_derivative=(n_wat2-1.)*(1.+2./3.*(n_wat2+2.)*(n_wat/3.-1/3./n_wat)**2)

    return n_density_derivative




import numpy as np
cimport numpy as np
from libc.math cimport isnan

cdef class CLUT:

    def __init__(self, A, axes=None, debug=False):
        '''
        A: N-dim array
        axes: list of axes. each axis is either:
                * a list or array of floats (index lookup is activated)
                * None: index lookup is not activated
        '''
        cdef int i
        self.shape = np.array(list(A.shape)).astype('int32')
        self.ndim = A.ndim
        self.data = A.astype('float32').ravel(order='C')
        self.debug = debug

        self._index = np.zeros(A.ndim, dtype='int32')-999
        self._inf = np.zeros(A.ndim, dtype='int32')
        self._x = np.zeros(A.ndim, dtype='float32')
        self._dim_interp = np.zeros(A.ndim, dtype='int32')
        self._interp = np.zeros(A.ndim, dtype='int32')
        self.scaling = np.zeros(A.ndim, dtype='float32')
        self.reverse = np.zeros(A.ndim, dtype='int32')
        self.bounds = np.zeros((A.ndim, 2), dtype='float32')

        self.dim_has_axis = np.zeros(A.ndim, dtype='int32')
        if axes is None:
            axes = [None] * A.ndim

        max_axis_size = 1
        Ninv = 1
        Nmin = 3.   # at least Nmin bins per element
        for a in axes:
            if a is None:
                continue
            if (len(a) > max_axis_size):
                max_axis_size = len(a)
            deltamin = np.amin(np.abs(np.diff(a)))/float(abs(a[0] - a[-1]))
            if int(Nmin/deltamin) > Ninv:
                Ninv = int(Nmin/deltamin)

        # set up the axes
        # and the inverted axes for faster inversion
        ax = np.zeros((A.ndim, max_axis_size), dtype='float32')+np.NaN
        iax = np.zeros((A.ndim, Ninv), dtype='int64') - 999
        for i, a in enumerate(axes):
            if a is None:
                continue
            assert isinstance(a, (np.ndarray, list))
            if len(a) != self.shape[i]:
                raise Exception('CLUT: shape mismatch for axis {}'.format(i))
            ax[i,:len(a)] = a
            self.dim_has_axis[i] = 1

            self.reverse[i] = a[-1] < a[0] # whether axis is reversed

            # if axis sorted in decreasing order, sort it in ascending order first
            if self.reverse[i]:
                a = a[::-1]

            self.bounds[i,0] = a[0]
            self.bounds[i,1] = a[-1]

            # inverted axis
            v = np.linspace(a[0], a[-1], Ninv+1, dtype='float64')[:-1]
            iax[i, :] = np.searchsorted(a, v)-1
            iax[i, 0] = 0.
            # set the last item of each consecutive series to -1
            # (uncertain bracketing)
            iax[i,:-1][np.diff(iax[i,:]) != 0] = -1

            # if self.reverse[i]:
                # iax[i,iax[i,:]>=0] = len(a) - 2 - iax[i,iax[i,:]>=0]

            # scaling factor for quick index lookup
            self.scaling[i] = Ninv/float(a[-1]-a[0])

        self.axes = ax
        self.invax = iax


    cdef float get(self, int[:] x):
        '''
        Get array value at integer coordinates x
        '''
        cdef int index = x[0]
        cdef int i = 0

        # row-major (C): last dimension is contiguous in memory
        for i in range(1, self.ndim):
            index *= self.shape[i]
            index += x[i]

        return self.data[index]


    cdef set(self, float value, int[:] x):
        '''
        Set value at integer coordinates x
        '''
        cdef int index = x[0]
        cdef int i = 0

        # row-major (C): last dimension is contiguous in memory
        for i in range(1, self.ndim):
            index *= self.shape[i]
            index += x[i]

        self.data[index] = value


    cdef int index(self, int i, int j):
        '''
        set current index on dimension i using integer indexing
        (no interpolation)
        '''
        self._inf[i] = j
        self._interp[i] = 0

        return 0


    cdef int indexf(self, int i, float x):
        '''
        set current index of dimension i using floating index
        (interpolation)
        '''
        self._inf[i] = <int>x
        self._x[i] = x - self._inf[i]
        self._interp[i] = 1

        return 0


    cdef int lookup(self, int i, float v) except -999:
        '''
        index lookup for axis i with value v:
        sets up index j such that v[j] < v and interpolation ratio

        returns:
            0 on success
            -1 in case of lower out-of-bounds
            1 in case of upper out-of-bounds

        in case of out-of-bounds, set up the axes by clipping
        '''
        cdef long int j, jj
        cdef float lower, upper

        if isnan(v):
            raise Exception('lookup of NaN')

        if not self.dim_has_axis[i]:
            raise Exception('Trying to use index lookup without associated axis')

        if not self.reverse[i]:
            # lower end clipping
            if v < self.bounds[i,0]:
                self._inf[i] = 0
                self._interp[i] = 0
                return -1
            # higher end clipping
            if v > self.bounds[i,1]:
                self._inf[i] = self.shape[i]-1
                self._interp[i] = 0
                return 1
        else:
            # lower end clipping
            if v > self.bounds[i,1]:
                self._inf[i] = 0
                self._interp[i] = 0
                return -1
            # higher end clipping
            if v < self.bounds[i,0]:
                self._inf[i] = self.shape[i]-1
                self._interp[i] = 0
                return 1

        self._interp[i] = 1

        # index in the lookup array
        j = <long int>((v - self.bounds[i,0])*self.scaling[i])
        if (j < 0) or (j > self.invax.shape[1]):
            raise Exception('Index error in lookup: index={}, value={}'.format(j, v))
        if j == self.invax.shape[1]:
            j -= 1

        # index in the array
        jj = self.invax[i,j]

        if (jj < 0):
            if v > self.axes[i, self.invax[i, j+1]]:
                jj = self.invax[i, j+1]
            else:
                jj = self.invax[i, j-1]

        if self.reverse[i]:
            jj = self.shape[i] - 2 - jj

        if self.debug:
            # verifications
            if jj<0:
                raise Exception('Consistency error in lookup, jj={}'.format(jj))
            if not ((self.axes[i,jj] - v) * (self.axes[i,jj+1] - v) < 0):
                raise Exception('Could not verify {} between {} and {}'.format(
                    v, self.axes[i,jj], self.axes[i,jj+1]))

        self._inf[i] = jj
        if self._inf[i] < 0:
            raise Exception('Error: negative index on axis {} (index is {}, value is {})'.format(
                i, self._inf[i], v))

        lower = self.axes[i, self._inf[i]]
        upper = self.axes[i, self._inf[i]+1]
        self._x[i] = (v - lower)/(upper - lower)

        return 0

    cdef float interp(self):
        '''
        Interpolate a value in the array at floating point coordinates x
            - if an axis is provided for a given dimension i, performs index lookup
            - otherwise, interpolates using floating-point index
            - if coordinate is < 0, use int index[i]
        '''
        cdef float coef
        cdef float rvalue = 0.
        cdef int j, d, b, D
        cdef int n_dim_interp = 0

        for j in range(self.ndim):
            if self._interp[j]:
                self._dim_interp[n_dim_interp] = j
                n_dim_interp += 1
            else:
                self._index[j] = self._inf[j]

        # loop over the 2^n dimensions to interpolate
        for j in range(1<<n_dim_interp):

            # calculate the weight of the current item
            coef = 1.
            for d in range(n_dim_interp):
                # number of the current dimension
                D = self._dim_interp[d]

                # b is the value of the 'd'th bit in j (ie corresponding to
                # interpolation dimension number d)
                # it is used to determine if, in the current combination j, the
                # 'd'th dimension is a 'inf' or a 'inf+1'='sup' so as to
                # determine if coef has to be multiplied by x or 1-x
                b = (j & (1<<d))>>d

                self._index[D] = self._inf[D] + b

                if b:
                    coef *= self._x[D]
                else:
                    coef *= 1 - self._x[D]

            rvalue += coef * self.get(self._index)

        return rvalue


def test_get():
    cdef CLUT A = CLUT(np.array([[1, 2, 3], [4, 5, 6]]), axes=[None, None])

    assert A.get(np.array([0, 0], dtype='int32')) == 1
    assert A.get(np.array([1, 1], dtype='int32')) == 5

def test_lookup():

    cdef CLUT A = CLUT(np.array([[1, 2, 3], [4, 5, 6]]), axes=[[1, 2], [10, 11, 12]])

    assert A.lookup(0, 1.5) == 0
    assert A.lookup(0, 0.5) < 0
    assert A.lookup(0, 2.5) > 0

    assert A.lookup(1, 11.5) == 0
    assert A.lookup(1, 0.) < 0
    assert A.lookup(1, 100.) > 0

def test_interp():
    cdef CLUT A = CLUT(np.array([[1, 2, 3], [4, 5, 6]]), axes=[[1, 2], [10, 11, 12]])

    assert A.lookup(0, 1.5) == 0
    assert A.lookup(1, 11.5) == 0

    print A.interp()
    assert A.interp() == 4

def test():
    '''
    test this module
    '''
    test_get()
    test_interp()




import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from luts.luts import Idx

def Get_3Dcells_indices(NX, NY, NZ):
    '''
    set up a rectangular regular 3D grid indices
    
    Inputs:
        Number of grid cells in each dimension
        
    Ouputs:
        triplet of 3D indices
    '''
    Ncell = NX*NY*NZ
    # from cell number to x,y and z indices
    return np.unravel_index(np.arange(Ncell, dtype=np.int32), (NX, NY, NZ), order='C')


def Get_3Dcells_neighbours(NX, NY, NZ, BOUNDARY_ABS=-5, periodic=False,
                          BOUNDARY_BOA=-2, BOUNDARY_TOA=-1):
    '''
    Computes the 3D neighbouring cells indices, one for each of the 6 cube faces
    
    Inputs:
        Number of grid cells in each dimension
        
    Keyword:
        - periodic : the neighbours are horizontally periodic, otherwise it is an absorbing boundary
        
    Outputs:
        2D array (6, Ncell) containing the neighbouring cell indices for each of the 6 cuboid faces,
            with the convention order, +X,-X,+Y,-Y,+Z,-Z

    '''
    idx, idy, idz  = Get_3Dcells_indices(NX, NY, NZ)
    # indices of neighbouring cells in rectangular grid
    neigh_idx      = np.vstack((idx+1, idx-1, idx  , idx  , idx  , idx  )) # by convention POSITIVE first
    neigh_idy      = np.vstack((idy  , idy  , idy+1, idy-1, idy  , idy  ))
    neigh_idz      = np.vstack((idz  , idz  , idz  , idz  , idz+1, idz-1))
    
    if periodic: 
        neigh = np.ravel_multi_index((neigh_idx, neigh_idy, neigh_idz), 
                                       dims=(NX, NY, NZ), mode=('wrap','wrap','clip'))
    else:
        neigh = np.ravel_multi_index((neigh_idx, neigh_idy, neigh_idz), 
                                       dims=(NX, NY, NZ), mode=('clip','clip','clip'))
    ## boundaries neighbouring
    # with 'clip' mode, if outside the domain then the neighbour index is the same as the cell index
    neigh[np.equal(neigh , np.arange(NX*NY*NZ, dtype=np.int32))] = BOUNDARY_ABS 
    # by definition -Z neighbour at the domain boundary is BOA
    neigh[5, np.where(neigh[5,:]==BOUNDARY_ABS)] = BOUNDARY_BOA 
    # by definition +Z neighbour at the domain boundary is TOA
    neigh[4, np.where(neigh[4,:]==BOUNDARY_ABS)] = BOUNDARY_TOA
    # by convention +Z neighbour at the domain boundary -1 is also TOA
    neigh[4, np.where(neigh[4,:]%NZ==        0)] = BOUNDARY_TOA 
    
    return neigh.astype(np.int32)


def Get_3Dcells(Nx=1, Ny=1, Nz=50, Dx=1., Dy=1., Dz=1.,x=None, y=None, z=None, periodic=False,
                HORIZ_EXTENT_LENGTH=0, SAT_ALTITUDE=1e3):
    '''
    return the cells geometrical properties for use in 3D atmospheric profile object

    Keywords:
        - Nx, Ny and Nz are the number of cells in each dimension
        - Dx, Dy and Dz are the cells dimensions in km (default 1.)
        - x(Nx), y(Ny), z(Nz), coordinates can be provided instead, it erase Ni and Di
        - HORIZ_EXTENT_LENGTH: in km, is not 0, then one cell before and one after in X and Y 
            are added with a specific length of HORIZ_EXTENT_LENGTH. it results in a total
            number of cells being NX = Nx+2 and NY = Ny+2
        - SAT_ALTITUDE: Max altitude in km for sensors location within the 3D grid
        - periodic : the neighbours are horizontally periodic, otherwise it is an absorbing boundary
    '''
    # CELLS INDEXING
    DN             = 2 if HORIZ_EXTENT_LENGTH !=0 else 0
    sl             = slice(DN//2, -DN//2) if DN==2 else slice(None, None)
        
    if x is None:
        NX   = Nx + DN  # Number of cells in x
        # cells boundaries coordinates 
        # horizontal 
        Hx   = Nx*Dx/2.  # x central domain half length (km)
        x    = np.zeros((NX+1), dtype=np.float32)
        x[0] = -(Hx + HORIZ_EXTENT_LENGTH)
        x[-1]= (Hx  + HORIZ_EXTENT_LENGTH)
        x[sl]= np.linspace(-Hx, Hx, num=Nx+1)
    else:
        NX   = x.size-1
        
    if y is None:
        NY   = Ny + DN  # Number of cells in x
        # cells boundaries coordinates 
        # horizontal 
        Hy   = Ny*Dy/2.  # y central domain half length (km)
        y    = np.zeros((NY+1), dtype=np.float32)
        y[0] = -(Hy + HORIZ_EXTENT_LENGTH)
        y[-1]= (Hy  + HORIZ_EXTENT_LENGTH)
        y[sl]= np.linspace(-Hy, Hy, num=Ny+1)
    else:
        NY   = y.size-1
        
    # vertical boundaries
    if z is None :
        # we add a empty very thin cell above TOA (just for interfacing purposes)
        NZ    = Nz + 1
        z     = np.zeros((NZ+1), dtype=np.float32)
        z[:-1]= np.linspace(0, Nz*Dz, num=NZ)
        z[-1] = SAT_ALTITUDE # above TOA , sensor max level
    else:
        NZ    = z.size-1
    
    # from cell number to x,y and z indices
    idx, idy, idz  = Get_3Dcells_indices(NX, NY, NZ)
    # indices of neighbouring cells in rectangular grid
    neigh          = Get_3Dcells_neighbours(NX, NY, NZ, periodic=periodic)
    # Bounding boxes, lower left and upper right corners
    pmin           = np.zeros((3, NX*NY*NZ), dtype=np.float32)
    pmax           = np.zeros_like(pmin)
    pmin[0,:]      = x[idx]
    pmax[0,:]      = x[idx+1]
    pmin[1,:]      = y[idy]
    pmax[1,:]      = y[idy+1]
    pmin[2,:]      = z[idz] # ! from bottom to top
    pmax[2,:]      = z[idz+1]

    return (idx,idy,idz), (NX,NY,NZ), (x,y,z), neigh, pmin, pmax


def locate_3Dregular_cells(xgrid,ygrid,zgrid,x,y,z):
    '''
    return the cells indices corresponding the the coordinates x,y,z
    in a regular grid whose limits are defined by xgrid,ygrid and zgrid
    '''
    return  np.ravel_multi_index(( \
            np.floor(interp1d(xgrid, np.arange(len(xgrid)))(x)).astype(int) ,
            np.floor(interp1d(ygrid, np.arange(len(ygrid)))(y)).astype(int) ,
            np.floor(interp1d(zgrid, np.arange(len(zgrid)))(z)).astype(int)),
                     dims = (len(xgrid)-1, len(ygrid)-1, len(zgrid)-1))


def satellite_view(mlut, Nx, Ny, x0, y0, wl, interp_name='none',
                   color_bar='Blues_r', fig_size=(8,8), font_size=int(18),
                   vmin = None, vmax = None, scale=False, save_file=None):
    """
    Description: The function give a 'satellite' 2D image of the SMART-G 3D atm return results

    ===Parameters:
    mlut        : SMART-G return MLUT object
    Nx, Ny      : Number of sensors in x and y axis
    x0, y0      : x and y arrays with x and y cell boundary positions of the domain
    wl          : Wavelength
    interp_name : Interpolations for imshow/matshow, i.e. nearest, bilinear, bicubic, ...
    color_bar   : Bar color of the figure, i.g 'Blues_r', 'jet', ...
    fig_size    : A tuple with width and height of the figure in inches
    vmin, vmax  : Color bar interval
    scale       : If True scale between 0 and 1 (or between vmin and vmax if not None)
    font_size   : Font size of the figure
    save_file   : If not None, save the generated image in pdf, i.g. save_file = 'test', save as 'test.pdf'
    """
    # Force the use of only 'I_up (TOA), TODO add the option to choose between I, Q, U, V
    stokes_name = 'I_up (TOA)'

    # First check the Azimuth and Zenith angles dimensions exist, if not the case return an error message
    if ("Azimuth angles" or "Zenith angles") not in mlut[stokes_name].names:
        raise NameError("The Azimuth angles and/or Zenith angles dimension(s) are/is missing")

    # The number of dimensions in the LUT
    axis_number = len(mlut[stokes_name].names)
    
    ind = [slice(None)]*axis_number # if axis_number = 3, tuple(ind) equivalent to [:,:,:]

    # Two last indices for axis Azimuth angles and Zenith angles forced to 0 (consider we have only one sun position)
    ind[-1] = 0; ind[-2] = 0 # TODO Consider also the case where several sun position are given

    # if Azimuth dim or Zenith dim > 1 -> return an error message. TODO to remove once the option above is added
    if ((mlut.axes["Azimuth angles"].size or mlut.axes["Zenith angles"].size) > 1):
        raise NameError("Dimension size > 1 is not authorized for both Azimuth and Zenith angles")

    # If we have the wavelength dimension
    if "wavelength" in mlut[stokes_name].names:
        ind[-3] = Idx(wl) # TODO Enable a default value, for example for the monochromatique case


    # Check if the product of Nx and Ny is equal to the number of sensors
    if "sensor index" in mlut[stokes_name].names:
        sensor_number = mlut.axes["sensor index"].size
    else:
        sensor_number = int(1)
    if (Nx*Ny != sensor_number):
        raise NameError("The product of Nx and Ny must be equal to the number of sensors!")


    # Convert the 1D results to a 2D matrix. The order of Nx and Ny below is very important!
    matrix = mlut[stokes_name][tuple(ind)].reshape(Ny,Nx)
    # The variable matrix is now in the following form:
    # - x0 ... xn
    # y0
    #  :
    # yn

    # By default in the imshow function, the origin (origin='upper') i.e matrix[0,0] is at the upper left,
    # and we want the origin at bottom left (origin='lower).
    plt.figure(figsize=fig_size)
    plt.rcParams.update({'font.size':font_size})

    # Deal with all the possibilties where vmin, vmax and scale are used
    if scale:
        vmin_scale = 0.; vmax_scale = 1.
        if vmin is not None : vmin_scale = vmin
        if vmax is not None : vmax_scale = vmax
        matrix = np.interp(matrix, (matrix.min(), matrix.max()), (vmin_scale, vmax_scale))

    # Now call the function imshow for the 2D drawing
    img = plt.imshow(matrix, vmin=vmin, vmax=vmax, origin='lower', cmap=plt.get_cmap(color_bar),
                     interpolation=interp_name, extent=[x0.min(),x0.max(),y0.min(),y0.max()])

    # Color bar parametrization
    cbar = plt.colorbar(img, shrink=0.7, orientation='vertical')
    cbar.set_label(r''+ stokes_name, fontsize = font_size)

    # Labels for x and y axes
    plt.xlabel(r'X (km)')
    plt.ylabel(r'Y (km)')

    # If the option save_file is used, save the file in pdf
    if (save_file is not None):
        # Deal with the case where we have not specified the '.pdf' at the end
        if not save_file.endswith('.pdf'):
            save_file += '.pdf'
        plt.savefig(save_file)